/*
	ICP DAS Virtual COM driver.

        Author: Golden Wang

        This program is free software; you can redistribute it and/or modify
        it under the terms of the GNU General Public License as published by
        the Free Software Foundation; either version 2, or (at your option)
        any later version.

        This program is distributed in the hope that it will be useful,
        but WITHOUT ANY WARRANTY; without even the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
        GNU General Public License for more details.

        You should have received a copy of the GNU General Public License
        along with this program; if not, write to the Free Software Foundation,
        Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* File level history (record changes for this file here.)
 *  
 *      v 0.0.0 17 Dec 2010 by Golden Wang
 *            Linux Virtual COM Driver for PDS Series.
 *     
 *     
*/

#include "ixvcom.h"
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/netdevice.h>
#include <linux/version.h>
#include <linux/init.h>
//#include <linux/autoconf.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/io.h>
#include <linux/mm.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
#include <linux/smp_lock.h>
#endif
#include <linux/proc_fs.h>
#include <linux/delay.h>
#include <linux/poll.h>

#define DRV_NAME  "icpdas-ixvcom"
#define IXVCOM_EVENT_TXLOW              1
#define IXVCOM_EVENT_HANGUP             2

#define SERIAL_DO_RESTART
#define SERIAL_TYPE_NORMAL              1
#define SERIAL_TYPE_CALLOUT             2

#define IXVCOM_VCOM_COMMAND_SET         1
#define IXVCOM_LOCAL_COMMAND_SET        2

#define LOCAL_CMD_TTY_USED              1
#define LOCAL_CMD_TTY_UNUSED            2
#define IXVCOM_NET_CONNECTED            3
#define IXVCOM_NET_DISCONNECTED         4
#define IXVCOM_NET_SETTING              5
#define IXVCOM_NET_GET_TTY_STATUS       6

#define IXVCOM_CMD_TIMEOUT              1*HZ/40  // 25 milliseconds
//#define IXVCOM_CMD_TIMEOUT              1*HZ  // 1 seconds
#define IXVCOM_CMD_TIMEOUT2             1*HZ/40  // 25 milliseconds
//#define IXVCOM_CMD_TIMEOUT2             1*HZ  // 1 seconds

#define IXVCOM_NET_CMD_RETRIEVE         1
#define IXVCOM_NET_CMD_RESPONSE         2

#define IXVCOM_NET_DEV_OPENED           0x01
#define IXVCOM_NET_DEV_CONNECTED        0x02
#define IXVCOM_NET_DEV_DISCONNECTED     0x04
#define IXVCOM_NET_DO_SESSION_RECOVERY  0x08
#define IXVCOM_NET_DO_INITIALIZE        0x10
#define IXVCOM_NET_TTY_INUSED           0x20

//VxCOMM Command
#define IXVCOMCMD		   1				
#define VCOM_CMD_SET_BAUD	   2
#define VCOM_CMD_SET_DATAFORMAT    3
#define VCOM_CMD_SET_LINECTRL      5
#define VCOM_CMD_GET_MODULE_NAME  10
#define VCOM_CMD_SET_BREAK        27
#define VCOM_CMD_SET_FOLLOWCTRL   28
#define VCOM_CMD_FLUSH_TXRXBUFFER 35


#ifndef B921600
#define B921600 (B460800 + 1)
#endif

#ifndef MAX_SCHEDULE_TIMEOUT
#define MAX_SCHEDULE_TIMEOUT    ((long)(~0UL>>1))
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,31))
#define MX_SESSION()    task_session_vnr(current)
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
#define MX_SESSION()    task_session_nr(current)
#else
#define MX_SESSION()    (pid_t)tty->session
#endif
#else
#define MX_SESSION()    current->signal->session
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
#define MX_CGRP()   task_pgrp_nr(current)
#else
#define MX_CGRP()   process_group(current)
#endif

#ifndef MIN
#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#endif

#ifndef ASYNC_CALLOUT_ACTIVE
#define ASYNC_CALLOUT_ACTIVE 0
#endif

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,36))
/* include/linux/semaphore.h modification */
#define init_MUTEX(sem) sema_init(sem, 1)
#endif

MODULE_AUTHOR("Golden <service@icpdas.com>");
MODULE_DESCRIPTION("Virtual COM driver for ICP DAS PDS Series");
MODULE_SUPPORTED_DEVICE("ICP DAS PDS Serires");
MODULE_LICENSE("GPL v2");

struct server_setting_struct
{
	int32_t server_type;
	int32_t disable_fifo;
	int32_t com_port;
};

struct ixvcom_struct 
{
	int server_type;
	int port;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
	struct tty_port ttyPort;
#endif
	//int base;               /* port base address */
	//int irq;                /* port using irq no. */
	//char irq_name[16];      /* irq_name for /proc/interrupt */
	//int irq_request;        /* check the status of "request_irq" */
	//int vector;             /* port irq vector */
	//int vectormask;         /* port vector mask */
	int rx_high_water;
	int rx_trigger;         /* Rx fifo trigger level */
	int rx_low_water;
	int baud_base;          /* max. speed */
	int flags;              /* defined in tty.h */
	int uart_type;               /* UART type */
	struct tty_struct *tty;
	int read_status_mask;
	int ignore_status_mask;
	int xmit_fifo_size;
	int custom_divisor;
	int x_char;             /* xon/xoff character */
	int close_delay;
	unsigned short closing_wait;
	int IER;                /* Interrupt Enable Register */
	int MCR;                /* Modem control register */
	int ACR;
	unsigned long event;
	int count;              /* # of fd on device */
	int blocked_open;       /* # of blocked opens */
	int modem_control;  /* Modem control register */
	int modem_status;   /* Line status */

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18)
	pid_t session;    /* Session of opening process */
	pid_t pgrp;           /* pgrp of opening process */
#else
	long session;           /* Session of opening process */
	long pgrp;
#endif

	unsigned char *xmit_buf;
	int xmit_head;
	int xmit_tail;
	int xmit_cnt;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0))
	struct tq_struct    tqueue;
	struct tq_struct    process_flip_tqueue;
#else
	struct work_struct  tqueue;
	struct work_struct  process_flip_tqueue;
#endif


#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,18)
	struct termios normal_termios;
	struct termios callout_termios;
#else
	struct ktermios normal_termios;
	struct ktermios callout_termios;
#endif
	wait_queue_head_t open_wait;
	wait_queue_head_t close_wait;
	wait_queue_head_t delta_msr_wait;
	struct async_icount icount;     /* kernel counters for the 4 input interrupts */
	struct ixvcom_hwconf *net_dev;

/*
	int timeout;
	int MaxCanSetBaudRate;
	int opmode_ioaddr;
	unsigned char stop_rx;
	unsigned char ldisc_stop_rx;
	long realbaud;
	unsigned char err_shadow;
	spinlock_t slock;
*/
	spinlock_t          tx_lock;
	struct semaphore    rx_semaphore;
};

struct ixvcom_hwconf
{
	int server_type;
	wait_queue_head_t   initialize_wait;
	wait_queue_head_t   select_in_wait;
	wait_queue_head_t   select_out_wait;
	wait_queue_head_t   select_ex_wait;
	wait_queue_head_t   cmd_rsp_wait;
	int                 tx_ready;
	int                 rx_ready;
	int                 cmd_ready;
	int                 wait_oqueue_responsed;
	int                 oqueue;
	unsigned char       cmd_buffer[84];
	unsigned char       rsp_buffer[84];
	struct semaphore    cmd_semaphore;
	int                 rsp_length;
	unsigned long               flag;
	struct proc_dir_entry *node_entry;
	struct ixvcom_struct  *tty_node;
	struct semaphore    semaphore;
	int                 do_session_recovery_len;
};

//static char *ixvcom_version = "ixvcom-0.0.0";
//static char *ixvcom_name = "ICP DAS Virtual COM Driver";
static struct tty_driver *ixvcom_driver;

static struct ixvcom_struct ixvcom_table[VXCOM_MAX_PORTS];
static struct ixvcom_hwconf ixvcom_net_dev[VXCOM_MAX_PORTS];
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0))
static struct tty_struct *ixvcom_tty[VXCOM_MAX_PORTS];

#if (LINUX_VERSION_CODE > KERNEL_VERSION(3,2,20))
static struct ktermios *ixvcom_termios[VXCOM_MAX_PORTS];
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,20))
static struct ktermios *ixvcom_termios[VXCOM_MAX_PORTS];
static struct ktermios *ixvcom_termios_locked[VXCOM_MAX_PORTS];
#else
static struct termios *ixvcom_termios[VXCOM_MAX_PORTS];
static struct termios *ixvcom_termios_locked[VXCOM_MAX_PORTS];
#endif
#endif

int 				ixvcom_major;
static struct proc_dir_entry *  ixvcom_proc_dir;

static unsigned char *          ixvcom_tmp_buf;
static struct semaphore         ixvcom_tmp_buf_sem;

static struct file_operations ixvcom_net_fops;

static int  ixvcom_init_tty(void);	


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static void ixvcom_do_softint(void *);
static void ixvcom_flush_to_ldisc(void *);
#else
static void ixvcom_do_softint(struct work_struct *work);	
static void ixvcom_flush_to_ldisc(struct work_struct *work);
#endif

static int  ixvcom_open(struct tty_struct *,struct file *);	
static void ixvcom_close(struct tty_struct *,struct file *);	

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10))
static int ixvcom_write(struct tty_struct *,int,const unsigned char *,int);
#else
static int ixvcom_write(struct tty_struct *,const unsigned char *,int);
#endif

static int  ixvcom_write_room(struct tty_struct *);
static void ixvcom_flush_buffer(struct tty_struct *);	
static void ixvcom_ldisc_flush_buffer(struct tty_struct *);
static int  ixvcom_chars_in_buffer(struct tty_struct *);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26))
static void ixvcom_put_char(struct tty_struct *,unsigned char);
#else
static int  ixvcom_put_char(struct tty_struct *,unsigned char);
#endif

static void ixvcom_flush_chars(struct tty_struct *);
static void ixvcom_throttle(struct tty_struct *);
static void ixvcom_unthrottle(struct tty_struct *);	

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static void ixvcom_set_termios(struct tty_struct *,struct termios *);
static int  ixvcom_port_init(struct ixvcom_struct *,struct termios *);	
#else
static void ixvcom_set_termios(struct tty_struct *,struct ktermios *);
static int  ixvcom_port_init(struct ixvcom_struct *,struct ktermios *);	
#endif

static void ixvcom_stop(struct tty_struct *);
static void ixvcom_start(struct tty_struct *);
static void ixvcom_hangup(struct tty_struct *);
static inline void ixvcom_check_modem_status(struct ixvcom_struct *,int);
static int  ixvcom_block_til_ready(struct tty_struct *,struct file *, struct ixvcom_struct *);	
static int  ixvcom_startup(struct ixvcom_struct *,struct file *, struct tty_struct *);	
static void ixvcom_shutdown(struct ixvcom_struct *);	
static int  ixvcom_port_shutdown(struct ixvcom_struct *);	
static int  ixvcom_get_serial_info(struct ixvcom_struct *, struct serial_struct *);
static int  ixvcom_set_serial_info(struct ixvcom_struct *, struct serial_struct *);
static int  ixvcom_get_lsr_info(struct ixvcom_struct *,unsigned int *);
static void ixvcom_send_break(struct ixvcom_struct *,int);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39))
static int  ixvcom_ioctl(struct tty_struct *, uint, ulong);
static int  ixvcom_tiocmget(struct tty_struct *);
static int  ixvcom_tiocmset(struct tty_struct *, unsigned int, unsigned int);
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static int  ixvcom_ioctl(struct tty_struct *, struct file *, uint, ulong);
static int  ixvcom_tiocmget(struct tty_struct *, struct file *);
static int  ixvcom_tiocmset(struct tty_struct *, struct file *, unsigned int, unsigned int);
#else
static int  ixvcom_get_modem_info(struct ixvcom_struct *,unsigned int *);
static int  ixvcom_set_modem_info(struct ixvcom_struct *,unsigned int, unsigned int *);
#endif

//static void ixvcom_process_notify(struct ixvcom_hwconf *,char *,int);
//static void ixvcom_do_session_recovery(struct ixvcom_struct *);
static void ixvcom_do_session_recovery_data_format(struct ixvcom_struct *);
static void ixvcom_do_session_recovery_baud(struct ixvcom_struct *);
static void ixvcom_wait_until_sent(struct tty_struct *,int);
static int  ixvcom_wait_and_set_command(struct ixvcom_hwconf *,char,char);	
static int  ixvcom_wait_command_completed(struct ixvcom_hwconf *,char,char, long,char *,int *);	//done
static long ixvcom_wait_oqueue(struct ixvcom_struct *,long);	
static int  ixvcom_linectrl(struct ixvcom_hwconf *, int port, int modem_control);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
static int  ixvcom_break(struct tty_struct * ttyinfo, int break_state);
#else
static void ixvcom_break(struct tty_struct * ttyinfo, int break_state);
#endif

static void ixvcom_start_break(struct ixvcom_hwconf *net_dev, int port);
static void ixvcom_stop_break(struct ixvcom_hwconf *net_dev, int port);
//static int ixvcom_setxon_xoff(struct ixvcom_struct * info, int cmd);

/*
 * File operation declarations
 */

static int ixvcom_net_open(struct inode *, struct file * );

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
long ixvcom_net_ioctl(struct file *file, unsigned int, unsigned long);
#else
static int ixvcom_net_ioctl(struct inode *, struct file *,unsigned int, unsigned long );
#endif

static int      ixvcom_net_close(struct inode *,struct file * );
static ssize_t  ixvcom_net_read (struct file *file,char *buf,size_t count, loff_t *ppos);
static ssize_t  ixvcom_net_write(struct file *file,const char *buf, size_t count,loff_t *ppos);
static unsigned int ixvcom_net_select(struct file *file, struct poll_table_struct *);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
static int ixvcom_proc_show(struct seq_file *m, void *v)
{
        /* export major number */

        seq_printf(m, "maj: %d\n", ixvcom_major);

        return  0;
}

static int ixvcom_proc_open(struct inode *inode, struct file *file)
{
        return single_open(file, ixvcom_proc_show, NULL);
}

static const struct file_operations ixvcom_proc_fops = {
        .open   = ixvcom_proc_open,
        .read   = seq_read,
        .release = single_release,
};
#endif


#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
static struct tty_operations ixvcom_ops =
{
        .open = ixvcom_open,
        .close = ixvcom_close,
        .write = ixvcom_write,
        .put_char = ixvcom_put_char,
        .flush_chars = ixvcom_flush_chars,
        .write_room = ixvcom_write_room,
        .chars_in_buffer = ixvcom_chars_in_buffer,
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
        .flush_buffer = ixvcom_ldisc_flush_buffer,
        .wait_until_sent = ixvcom_wait_until_sent,
        .break_ctl = ixvcom_break,
#else
        .flush_buffer = ixvcom_flush_buffer,
#endif
        .ioctl = ixvcom_ioctl,
        .throttle = ixvcom_throttle,
        .unthrottle = ixvcom_unthrottle,
        .set_termios = ixvcom_set_termios,
        .stop = ixvcom_stop,
        .start = ixvcom_start,
        .hangup = ixvcom_hangup,
        .tiocmget = ixvcom_tiocmget,
        .tiocmset = ixvcom_tiocmset,
};
#endif

char AsciiToHex(char cASCII)
{
        if (cASCII > '9') return(cASCII-'A'+10);
        else return(cASCII-'0');
}

char HexToAscii(unsigned short hex)
{
        if (hex<10)return((char)(hex+'0'));
        else return((char)(hex-10+'A'));
}

#if ((LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) || ((LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15)) && defined(FEDORA)))
static void tty_buffer_free(struct tty_struct *tty, struct tty_buffer *b)
{
	if (b->size >= 512)
	{
		kfree(b);
	}
	else
	{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
                b->next = tty->port->buf.free;
                tty->port->buf.free = b;
#else
                b->next = tty->buf.free;
                tty->buf.free = b;
#endif
	}
}
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0))
static int ixvcom_proc_output(char *buf)
{
	char *p;

	p = buf;

        /* export major number */

	p += sprintf(p, "maj: %d\n", ixvcom_major);

	return  p - buf;
}


static int ixvcom_read_proc(char *page, char **start, off_t off, int count, int *eof, void *data)
{
	int len = ixvcom_proc_output (page);

	if (len <= off+count) 
		*eof = 1;

	*start = page + off;
	len -= off;

        if (len>count)
		len = count;

        if (len<0)
		len = 0;

	return len;
}
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static void ixvcom_do_softint(void *private_)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)private_;
#else
static void ixvcom_do_softint(struct work_struct *work)
{
	struct ixvcom_struct * info = container_of(work, struct ixvcom_struct, tqueue);
#endif

	struct tty_struct * tty;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (info)
	{
		tty = info->tty;

		if (tty)
		{
			if ( test_and_clear_bit(IXVCOM_EVENT_TXLOW, &info->event) )
			{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
				if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc->ops->write_wakeup )
				{
					tty->ldisc->ops->write_wakeup(tty);
				}
#else
				if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) && tty->ldisc->write_wakeup )
				{
					tty->ldisc->write_wakeup(tty);
				}
#endif
				wake_up_interruptible(&tty->write_wait);
			}

			if ( test_and_clear_bit(IXVCOM_EVENT_HANGUP, &info->event) )
			{
				tty_hangup(tty);
			}
		}
	}
	
	//printk("SSS %s ... END SSS\n",__FUNCTION__);
}

static int ixvcom_net_open(struct inode *inode, struct file *file )
{
	struct ixvcom_hwconf *net_dev;
	int rtn = 0;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0))
        struct proc_dir_entry *de;
#endif

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	try_module_get(THIS_MODULE);

	if ( !capable(CAP_SYS_ADMIN) )
	{
		rtn = -EPERM;
		goto done;
	}

	if ( file->private_data )
	{
		rtn = -EINVAL;
		goto done;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0))
        net_dev = (struct ixvcom_hwconf *)PDE_DATA(inode);
#else
        de = PDE(inode);

        if (!de)
        {
                rtn = -ENXIO;
                goto done;
        }

        net_dev = (struct ixvcom_hwconf *)de->data;
#endif

	if ( !net_dev )
	{
		rtn = -ENXIO;
		goto done;
	}

	file->private_data = (void *)net_dev;

	down(&net_dev->semaphore);

	if ( net_dev->flag & IXVCOM_NET_DEV_OPENED)
	{
		rtn = -EBUSY;
		goto unlock;
	}

	net_dev->flag |= IXVCOM_NET_DEV_OPENED;
	net_dev->tx_ready = 0;
	net_dev->rx_ready = 1;
	net_dev->cmd_ready = 0;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
	//printk("SSS %s ... port : %d\n",__FUNCTION__, net_dev->tty_node->port);
        tty_register_device(ixvcom_driver, net_dev->tty_node->port, NULL);
#endif

unlock:

	up( &net_dev->semaphore );

	file->private_data = (void*)net_dev;

done:

	if ( rtn )
	{
		module_put(THIS_MODULE);
	}

	//printk("SSS %s ... END SSS\n",__FUNCTION__);
	return rtn;
}

static int ixvcom_net_close (struct inode *inode, struct file *file )
{
	struct ixvcom_hwconf *net_dev;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0))
        net_dev = (struct ixvcom_hwconf *)PDE_DATA(inode);
#else
	net_dev = (struct ixvcom_hwconf *)(file->private_data);
#endif

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if ( !net_dev )
	{
		goto done;
	}

	net_dev->flag  &= ~IXVCOM_NET_DEV_OPENED;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
	//tty_unregister_device(ixvcom_driver, (net_dev->tty_node->port));	//port must - 1
	tty_unregister_device(ixvcom_driver, (net_dev->tty_node->port - 1));	//port must - 1
#endif

done:
	file->private_data = NULL;

	module_put(THIS_MODULE);
	//printk("SSS %s ... END SSS\n",__FUNCTION__);

	return 0;
}

static unsigned int ixvcom_net_select (struct file *file, struct poll_table_struct *table)
{
	unsigned int retval = 0;
	struct ixvcom_hwconf *net_dev = (struct ixvcom_hwconf *)file->private_data;

	//this function will run with polling.
	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!net_dev)
	{
		printk("net_dev is null\n");
		return retval;
	}

	poll_wait(file, &net_dev->select_in_wait, table);
	poll_wait(file, &net_dev->select_out_wait, table);
	poll_wait(file, &net_dev->select_ex_wait, table);

	if ( net_dev->tx_ready )
	{
		//printk("SSS %s ... net_dev->tx_ready is true SSS\n",__FUNCTION__);
		retval |= POLLIN | POLLRDNORM;
	}

	if ( net_dev->rx_ready )
	{
		//printk("SSS %s ... net_dev->rx_ready is true SSS\n",__FUNCTION__);
		retval |= POLLOUT | POLLWRNORM;
	}

	if ( net_dev->cmd_ready )
	{
		//printk("SSS %s ... net_dev->cmd_ready is true ... CMD : %s SSS\n",__FUNCTION__, net_dev->cmd_buffer);
		retval |= POLLPRI;
	}
	//printk("SSS %s ... END SSS\n",__FUNCTION__);

	return retval;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
long ixvcom_net_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
#else
static int ixvcom_net_ioctl(struct inode *inode, struct file *file,unsigned int cmd, unsigned long arg)
#endif
{
	struct ixvcom_hwconf *net_dev  = file->private_data;
	int rtn = 0;
	int size,len;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);
	//printk("SSS %s ... 1 ... port : %d\n",__FUNCTION__, net_dev->tty_node->port);

	if ( !net_dev )
	{
		rtn = -ENXIO;
		goto done;
	}

	size = _IOC_SIZE(cmd);

	switch (_IOC_NR(cmd))
	{
		case IXVCOM_NET_CMD_RETRIEVE :
			//printk("SSS %s ... 1 ... IXVCOM_NET_CMD_RETRIEVE : %s SSS\n",__FUNCTION__, net_dev->cmd_buffer);

			if (!net_dev->cmd_ready)
			{
				rtn = -ENXIO;
				goto done;
			}

			len = strlen(net_dev->cmd_buffer) + 1;

			rtn = access_ok( VERIFY_WRITE, (void *)arg, len)?0:-EFAULT;
			//printk("SSS %s ... 2 ... IXVCOM_NET_CMD_RETRIEVE : %s SSS\n",__FUNCTION__, net_dev->cmd_buffer);

			if (rtn)
			{
				goto done;
			}

			//printk("SSS %s ... 3 ... IXVCOM_NET_CMD_RETRIEVE : %s SSS\n",__FUNCTION__, net_dev->cmd_buffer);

			if (copy_to_user( (void *)arg, (void *)net_dev->cmd_buffer,len ))
			{
				rtn = -EFAULT;
				goto done;
			}

			//printk("SSS %s ... 4 ... IXVCOM_NET_CMD_RETRIEVE : %s SSS\n",__FUNCTION__, net_dev->cmd_buffer);
			net_dev->cmd_buffer[0] = HexToAscii(0);
			rtn = len;
			net_dev->cmd_ready = 0;
			break;

		case IXVCOM_NET_CMD_RESPONSE :
		{
			unsigned char rsp_buffer[84];

			//printk("SSS %s ... 1 ... IXVCOM_NET_CMD_RESPONSE  SSS\n",__FUNCTION__);
			rtn = access_ok( VERIFY_READ,  (void *)arg, size )?0:-EFAULT;

			if ( rtn )
			{
				goto done;
			}

			//printk("SSS %s ... 2 ... IXVCOM_NET_CMD_RESPONSE  SSS\n",__FUNCTION__);
			if (copy_from_user( (void *)rsp_buffer, (void *)arg,size))
			{
				rtn = -EFAULT;
				goto done;
			}

			//printk("SSS %s ... 3 ... IXVCOM_NET_CMD_RESPONSE : %s SSS\n",__FUNCTION__, rsp_buffer);

			if (AsciiToHex(rsp_buffer[0]) == IXVCOM_LOCAL_COMMAND_SET)
			{
				//printk("SSS %s ... 4 ... IXVCOM_NET_CMD_RESPONSE(for LOCAL) : %s SSS\n",__FUNCTION__, rsp_buffer);
				down(&net_dev->cmd_semaphore);
				memcpy(net_dev->rsp_buffer, rsp_buffer, size);
				net_dev->rsp_length = size;
				up(&net_dev->cmd_semaphore);

				if ( waitqueue_active(&net_dev->cmd_rsp_wait))
					wake_up_interruptible(&net_dev->cmd_rsp_wait);

				break;
			}

			//printk("SSS %s ... 5 ... IXVCOM_NET_CMD_RESPONSE(for LOCAL) : %s SSS\n",__FUNCTION__, rsp_buffer);

			down(&net_dev->cmd_semaphore);
			memcpy(net_dev->rsp_buffer, rsp_buffer, size);
			net_dev->rsp_length = size;
			up(&net_dev->cmd_semaphore);

			if ( waitqueue_active(&net_dev->cmd_rsp_wait))
				wake_up_interruptible(&net_dev->cmd_rsp_wait);

			break;

		}

		case IXVCOM_NET_CONNECTED :
		{
			struct ixvcom_struct *info;
			
			//printk("XXX %s ... IXVCOM_NET_CONNECTED ... 1 XXX\n",__FUNCTION__);

			if (!(info=net_dev->tty_node))
				break;

			if (net_dev->flag & IXVCOM_NET_DEV_DISCONNECTED)
			{
				//printk("XXX %s ... IXVCOM_NET_CONNECTED ... 2 XXX\n",__FUNCTION__);
				net_dev->flag &= ~IXVCOM_NET_DEV_DISCONNECTED;
				net_dev->flag |= IXVCOM_NET_DEV_CONNECTED;
				ixvcom_do_session_recovery_data_format(info);
				ixvcom_do_session_recovery_baud(info);
			}

			//printk("XXX %s ... IXVCOM_NET_CONNECTED ...  3 XXX\n",__FUNCTION__);

			break;
		}

		case IXVCOM_NET_DISCONNECTED :
		{
			//printk("XXX %s ... IXVCOM_NET_DISCONNECTED XXX\n",__FUNCTION__);
			net_dev->flag &= ~IXVCOM_NET_DEV_CONNECTED;
			net_dev->flag |= IXVCOM_NET_DEV_DISCONNECTED;
        		net_dev->flag &= ~IXVCOM_NET_TTY_INUSED;  //clean inused flag

			if (waitqueue_active(&net_dev->cmd_rsp_wait))
			{
				net_dev->wait_oqueue_responsed = 1;
				wake_up_interruptible(&net_dev->cmd_rsp_wait);
			}

			break;
		}

		case IXVCOM_NET_GET_TTY_STATUS:
		{
			int status;
			//printk("XXX %s ... IXVCOM_NET_GET_TTY_STATUS ... start XXX\n",__FUNCTION__);

			if (size != sizeof (status))
				goto done;

			rtn = access_ok( VERIFY_READ, (void *)arg, size )? 0 : -EFAULT;

			if ( rtn )
			{
				goto done;
			}

			status = (net_dev->flag & IXVCOM_NET_TTY_INUSED) ? 1 : 0;

			if (copy_to_user( (void *)arg, (void *)&status,size ))
			{
				rtn = -EFAULT;
				goto done;
			}
			//printk("XXX %s ... IXVCOM_NET_GET_TTY_STATUS ... end XXX\n",__FUNCTION__);

			break;
		}

		case IXVCOM_NET_SETTING :
		{
			struct server_setting_struct settings;
			struct ixvcom_struct *info;

			//printk("XXX %s ... IXVCOM_NET_SETTING ... start XXX\n",__FUNCTION__);
			if (!(info = net_dev->tty_node))
				break;

			if (size != sizeof (struct server_setting_struct))
				goto done;

			rtn = access_ok( VERIFY_READ,  (void *)arg, size )? 0 : -EFAULT;

			if ( rtn )
			{
				goto done;
			}

			if (copy_from_user( (void *)&settings, (void *)arg,size))
			{
				rtn = -EFAULT;
				goto done;
			}

			if (settings.disable_fifo)
				info->uart_type = PORT_16450;
			else
				info->uart_type = PORT_16550A;

			if (info->uart_type == PORT_16550A)
			{
					info->xmit_fifo_size = 16;
			}
			else
			{
				info->xmit_fifo_size = 1;
			}

			info->port = settings.com_port;
			//printk("info->port : %d\n",info->port);
			//printk("XXX %s ... IXVCOM_NET_SETTING ... end XXX\n",__FUNCTION__);

			break;
		}

		default :
			break;
	}

done:
		//printk("SSS %s ... 2 ... port : %d\n",__FUNCTION__, net_dev->tty_node->port);
		//printk("SSS %s ... END SSS\n",__FUNCTION__);
		return rtn;
}

static ssize_t ixvcom_net_write (struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	struct ixvcom_hwconf *net_dev  = file->private_data;
	ssize_t  rtn = 0;
	int cnt;
	struct ixvcom_struct *info;
	struct tty_struct * tty;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if ( !buf )
	{
		rtn = count; 
		goto done;
	}

	if ( !net_dev )
	{
		rtn = count; 
		goto done;
	}

	if (!(info = (struct ixvcom_struct *)net_dev->tty_node))
	{
		rtn = count; 
		goto done;
	}

	if (info->flags & ASYNC_CLOSING)
	{
		rtn = count; 
		goto done;
	}

	down(&info->rx_semaphore);

	if (!(tty = info->tty))
	{
		rtn = count; 
		up(&info->rx_semaphore);
		goto done;
	}

	if (test_bit(TTY_IO_ERROR, &tty->flags))
	{
		rtn = count; 
		up(&info->rx_semaphore);
		goto done;
	}

	if ( !net_dev->rx_ready )
	{
		up(&info->rx_semaphore);
		printk("Port %d RX is not ready\n", info->port);
		goto done;
	}

#if ((LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,15)) && !defined(FEDORA))
	if ((cnt = MIN(count,TTY_FLIPBUF_SIZE-tty->flip.count)) <= 0)
	{
#else
	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0))
        if ((cnt = tty_buffer_request_room(tty->port, count)) <= 0)
	#else
        if ((cnt = tty_buffer_request_room(tty, count)) <= 0)
	#endif
        {

#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0))
		if (!tty->icanon || tty->canon_data)
		{
			if (!test_and_set_bit(TTY_THROTTLED,&tty->flags))
			{
				ixvcom_throttle(tty);
			}
		}
#endif

		up(&info->rx_semaphore);
		goto done;
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,8,0))
	if (!tty->icanon || tty->canon_data)
	{
		if ((cnt = MIN(cnt,(N_TTY_BUF_SIZE-1) - tty->read_cnt )) <= 0)
		{
			if (!test_and_set_bit(TTY_THROTTLED,&tty->flags))
			{
				ixvcom_throttle(tty);
			}

			up(&info->rx_semaphore);

			goto done;
		}
	}
#endif

#if ((LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,15)) && !defined(FEDORA))
	if (copy_from_user( tty->flip.char_buf_ptr, buf,cnt ))
	{
#else
	#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0))
        if ((count = tty_insert_flip_string(tty->port, (unsigned char *)buf, cnt)))
	#else
        if ((count = tty_insert_flip_string(tty, (unsigned char *)buf, cnt)))
	#endif
        {

		#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0))
                tty_flip_buffer_push(tty->port);
		#else
                tty_flip_buffer_push(tty);
		#endif
#endif
		rtn = count; /* throw it away*/
		up(&info->rx_semaphore);
		goto done;
	}

#if ((LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,15)) && !defined(FEDORA))
	tty->flip.count += cnt;
	rtn = cnt;
	tty->flip.char_buf_ptr += cnt;
	info->icount.rx += cnt;
	memset(tty->flip.flag_buf_ptr,TTY_NORMAL,cnt);
	tty->flip.flag_buf_ptr += cnt;
#endif

	up(&info->rx_semaphore);
	schedule_work(&info->process_flip_tqueue);

done:

	return rtn;
}

static ssize_t ixvcom_net_read (struct file *file, char *buf, size_t count, loff_t *ppos)
{
	struct ixvcom_hwconf *net_dev  = file->private_data;
	ssize_t  rtn = 0;
	int cnt;
	struct ixvcom_struct *info;
	unsigned long flags;
	struct tty_struct * tty;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if ( !net_dev )
	{
		rtn = -ENXIO;
		goto done;
	}

	if (!(info = (struct ixvcom_struct *)net_dev->tty_node))
	{
		rtn = -ENXIO;
		goto done;
	}

	tty = info->tty;

	if ( !tty )
	{
		rtn = -ENXIO;
		goto done;
	}

	if ( info->x_char )
	{
		rtn = 1;

		if (copy_to_user( buf, &info->x_char,rtn ))
		{
			rtn = -EFAULT;
			goto done;
		}

		info->x_char = 0;

		spin_lock_irqsave(&info->tx_lock, flags);
		info->icount.tx++;
		spin_unlock_irqrestore(&info->tx_lock, flags);

		goto done;
	}

	spin_lock_irqsave(&info->tx_lock, flags);

	if (!info->xmit_buf || info->xmit_cnt <= 0)
	{
		rtn = 0;
		spin_unlock_irqrestore(&info->tx_lock, flags);
		goto done;
	}

	spin_unlock_irqrestore(&info->tx_lock, flags);

	while (count)
	{
		cnt = MIN(count, MIN( info->xmit_cnt, SERIAL_XMIT_SIZE - info->xmit_tail));

		if ( cnt <= 0 )
			break;

		if (copy_to_user( buf+rtn,info->xmit_buf + info->xmit_tail,cnt ))
		{
			rtn = -EFAULT;
			goto done;
		}

		rtn += cnt;
		count -= cnt;
	
		spin_lock_irqsave(&info->tx_lock, flags);

		info->xmit_cnt -= cnt;
		info->xmit_tail += cnt;
		info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE - 1);
		info->icount.tx += cnt;

		spin_unlock_irqrestore(&info->tx_lock, flags);
	}

	if (info->xmit_cnt <= 0)
	{
		net_dev->tx_ready = 0;
	}
	else
	{
		net_dev->tx_ready = 1;

		if ( waitqueue_active(&net_dev->select_in_wait))
			wake_up_interruptible(&net_dev->select_in_wait);
	}

	if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
            tty->ldisc->ops->write_wakeup )
        (tty->ldisc->ops->write_wakeup)(tty);
#else
            tty->ldisc.write_wakeup )
        (tty->ldisc.write_wakeup)(tty);
#endif

	wake_up_interruptible(&tty->write_wait);

done:
	return rtn;
}



/*
static void ixvcom_process_notify(struct ixvcom_hwconf *net_dev, char *rsp_buffer, int rsp_length)
{
	int state;
	struct ixvcom_struct *info = net_dev->tty_node;

	printk("XXX %s ... start XXX\n",__FUNCTION__);

	if (!info)
		return;

	if (rsp_length != 5)
		return;

	if (rsp_buffer[2] & VCOM_NOTIFY_MSR_CHG)
	{
		state = 0;

		if (rsp_buffer[3] & 0x10)
			state |= UART_MSR_CTS;

		if (rsp_buffer[3] & 0x20)
			state |= UART_MSR_DSR;

		if (rsp_buffer[3] & 0x80)
			state |= UART_MSR_DCD;

		ixvcom_check_modem_status(info, state);
	}

	if (rsp_buffer[2] & VCOM_NOTIFY_BREAK)
	{
		struct tty_struct *tty;

		down (&info->rx_semaphore);

		if (!(tty= info->tty))
		{
			up (&info->rx_semaphore);
			return;
		}

		tty_insert_flip_char(tty, 0, TTY_BREAK);
		up (&info->rx_semaphore);

		info->icount.rx ++;
		info->icount.brk++;

		schedule_work(&info->process_flip_tqueue);

		if ( info->flags & ASYNC_SAK )
		{
			do_SAK(info->tty);
		}
	}

	if (rsp_buffer[2] & VCOM_NOTIFY_PARITY)
		info->icount.parity++;

	if (rsp_buffer[2] & VCOM_NOTIFY_FRAMING)
		info->icount.frame++;

	if ((rsp_buffer[2] & VCOM_NOTIFY_SW_OVERRUN) || (rsp_buffer[2] & VCOM_NOTIFY_HW_OVERRUN))
		info->icount.overrun++;
}
*/
/*
static void ixvcom_do_session_recovery(struct ixvcom_struct *info)
{
	return;
}
*/

static void ixvcom_do_session_recovery_data_format(struct ixvcom_struct *info)
{
	struct tty_struct *tty;
        struct ixvcom_hwconf *net_dev;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
        struct termios *termio;
#else
        struct ktermios *termio;
#endif
        char mode_str[16];
	int port = 0;

        //printk("SSS %s ... START SSS\n",__FUNCTION__);

	tty = info->tty;

	port = info->port;
        net_dev = info->net_dev;

        if ( !tty || !net_dev)
                return;

        if (!(net_dev->flag & IXVCOM_NET_TTY_INUSED))
                return;

        if (info->flags & ASYNC_INITIALIZED)
        {
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
                if (!(termio = &info->tty->termios))
#else
                if (!(termio = info->tty->termios))
#endif
                        return;
        }
        else
        {
                if (!(termio = &info->normal_termios))
                        return;
        }

        down (&net_dev->cmd_semaphore);

	if ((termio->c_cflag & CSIZE) == CS5)
        {
                sprintf(mode_str, "%d", 5);
        }
        else if ((termio->c_cflag & CSIZE) == CS6)
        {
                sprintf(mode_str, "%d", 6);
        }
        else if ((termio->c_cflag & CSIZE) == CS7)
        {
                sprintf(mode_str, "%d", 7);
        }
        else if ((termio->c_cflag & CSIZE) == CS8)
        {
                sprintf(mode_str, "%d", 8);
        }

	if (termio->c_cflag & PARENB)
        {
                if (termio->c_cflag & PARODD)
                {
                        sprintf(mode_str, "%s%c", mode_str, 'o');
                }
                else
                {
                        sprintf(mode_str, "%s%c", mode_str, 'e');
                }
        }
        else
        {
                sprintf(mode_str, "%s%c", mode_str, 'n');
        }
	
	if (termio->c_cflag & CSTOPB)
        {
                sprintf(mode_str, "%s%d", mode_str, 2);
        }
        else
        {
                sprintf(mode_str, "%s%d", mode_str, 1);
        }

	sprintf(net_dev->cmd_buffer, "%d%02d%d%s", IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_DATAFORMAT, port, mode_str);

        //printk("XXX %s ... CMD : %s(len %d)\n",__FUNCTION__, net_dev->cmd_buffer, strlen(net_dev->cmd_buffer));

	net_dev->do_session_recovery_len = strlen(net_dev->cmd_buffer) + 1;
        net_dev->flag |= IXVCOM_NET_DO_SESSION_RECOVERY;
        net_dev->cmd_ready = 1;
        up (&net_dev->cmd_semaphore);

        if ( waitqueue_active(&net_dev->select_ex_wait))
        {
                wake_up_interruptible(&net_dev->select_ex_wait );
        }

	return;
}

static void ixvcom_do_session_recovery_baud(struct ixvcom_struct *info)
{
	struct tty_struct *tty;
	struct ixvcom_hwconf *net_dev;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
	struct termios *termio;
#else
	struct ktermios *termio;
#endif
	int32_t baud;
	int port = 0;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	tty = info->tty;

	net_dev = info->net_dev;
	port = info->port;

	if ( !tty || !net_dev)
		return;

	if (!(net_dev->flag & IXVCOM_NET_TTY_INUSED))
		return;

	if (info->flags & ASYNC_INITIALIZED)
	{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
                if (!(termio = &info->tty->termios))
#else
                if (!(termio = info->tty->termios))
#endif
			return;
	}
	else
	{
		if (!(termio = &info->normal_termios))
			return;
	}

	down (&net_dev->cmd_semaphore);

	switch ( termio->c_cflag & (CBAUD|CBAUDEX))
	{
		case B921600:
			baud = 921600L;
			break;

		case B460800:
			baud = 460800;
			break;

		case B230400:
			baud = 230400L;
			break;

		case B115200:
			baud = 115200L;
			break;

		case B57600:
			baud = 57600L;
			break;

		case B38400:
			baud = 38400L;

			if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI )
			{
				baud = 57600L;
			}

			if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI )
			{
				baud = 115200L;
			}

#ifdef ASYNC_SPD_SHI
			if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			{
				baud = 230400L;
			}
#endif

#ifdef ASYNC_SPD_WARP
			if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			{
				baud = 460800L;
			}
#endif
			break;

		case B19200:
			baud = 19200L;
			break;

		case B9600:
			baud = 9600L;
			break;

		case B4800:
			baud = 4800L;
			break;

		case B2400:
			baud = 2400L;
			break;

		case B1800:
			baud = 1800L;
			break;

		case B1200:
			baud = 1200L;
			break;

		case B600:
			baud = 600L;
			break;

		case B300:
			baud = 300L;
			break;

		case B200:
			baud = 200L;
			break;

		case B150:
			baud = 150L;
			break;

		case B134:
			baud = 134L;
			break;

		case B110:
			baud = 110L;
			break;

		case B75:
			baud = 75L;
			break;

		case B50:
			baud = 50L;
			break;

		default:
			baud = 0;
	}

        sprintf(net_dev->cmd_buffer, "%d%02d%d%d", IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BAUD, port, baud);

        //printk("XXX %s ... CMD : %s(len %d)\n",__FUNCTION__, net_dev->cmd_buffer, strlen(net_dev->cmd_buffer));

	net_dev->do_session_recovery_len = strlen(net_dev->cmd_buffer) + 1;
	net_dev->flag |= IXVCOM_NET_DO_SESSION_RECOVERY;
	net_dev->cmd_ready = 1;
	up (&net_dev->cmd_semaphore);

	if ( waitqueue_active(&net_dev->select_ex_wait))
	{
		wake_up_interruptible(&net_dev->select_ex_wait );
	}

	return;
}



static int ixvcom_wait_and_set_command(struct ixvcom_hwconf *net_dev, char command_set, char command)
{
	unsigned long et;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);


	if ((command_set != IXVCOM_LOCAL_COMMAND_SET) && ((net_dev->flag & IXVCOM_NET_DO_SESSION_RECOVERY) || (net_dev->flag & IXVCOM_NET_DEV_DISCONNECTED)))
	{
		//printk("XXX %s ... 1 XXX\n",__FUNCTION__);
		return -1;
	}

	et = jiffies + IXVCOM_CMD_TIMEOUT;

	//printk("XXX %s ... 2 ... net_dev->cmd_buffer[0] : %d\n",__FUNCTION__, AsciiToHex(net_dev->cmd_buffer[0]));

	while (1)
	{
		down (&net_dev->cmd_semaphore);

		if (AsciiToHex(net_dev->cmd_buffer[0]) == 0)
		{
			if(command_set == IXVCOM_LOCAL_COMMAND_SET)
			{
				//printk("XXX %s ... 3 ... command_set == IXVCOM_LOCAL_COMMAND_SET\n", __FUNCTION__);
				sprintf(net_dev->cmd_buffer, "%d%d", command_set, command);
			}
			else
			{
				//printk("XXX %s ... 4 ... command_set == IXVCOM_VCOM_COMMAND_SET\n", __FUNCTION__);
				sprintf(net_dev->cmd_buffer, "%d%02d", command_set, command);
			}

			up(&net_dev->cmd_semaphore);
			//printk("SSS %s ... End(normal) SSS\n",__FUNCTION__);
			return 0;
		}
		else if ((jiffies >= et) || (signal_pending(current)))

		{ // timeout
			if(command_set == IXVCOM_LOCAL_COMMAND_SET)
                        {
                                //printk("XXX %s ... 5 ... command_set == IXVCOM_LOCAL_COMMAND_SET\n", __FUNCTION__);
				sprintf(net_dev->cmd_buffer, "%d%d", command_set, command);
                        }
                        else
                        {
                                //printk("XXX %s ... 6 ... command_set == IXVCOM_VCOM_COMMAND_SET\n", __FUNCTION__);
                                sprintf(net_dev->cmd_buffer, "%d%02d", command_set, command);
                        }

			up (&net_dev->cmd_semaphore);
			//printk("SSS %s ... End(timeout)  SSS\n",__FUNCTION__);
			return 0;
		}
		else
		{
			up(&net_dev->cmd_semaphore);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(1);
			//printk("SSS %s ... End(nothing) SSS\n",__FUNCTION__);
		}
	}
}

static int ixvcom_wait_command_completed(struct ixvcom_hwconf *net_dev, char command_set, char command, long timeout, char *rsp_buf, int *rsp_len)
{
	long et = 0;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);


	if ((command_set != IXVCOM_LOCAL_COMMAND_SET) && ((net_dev->flag & IXVCOM_NET_DO_SESSION_RECOVERY) || (net_dev->flag & IXVCOM_NET_DEV_DISCONNECTED)))
	{
		//printk("XXX %s ... start ... 1 XXX\n",__FUNCTION__);
		return -1;
	}
	
	if (*rsp_len <= 0)
		return -1;

	while (1)
	{
		down(&net_dev->cmd_semaphore);

		//printk("SSS %s ... rsp_buf : %s SSS\n",__FUNCTION__, net_dev->rsp_buffer);
		if (net_dev->rsp_length && (AsciiToHex(net_dev->rsp_buffer[0]) == command_set))
		{
			//printk("SSS %s ... 2 SSS\n",__FUNCTION__);
			if (net_dev->rsp_length > *rsp_len)
				return -1;

			*rsp_len = net_dev->rsp_length;
			memcpy(rsp_buf, net_dev->rsp_buffer, *rsp_len);
			net_dev->rsp_length = 0;
			up(&net_dev->cmd_semaphore);
			//printk("SSS %s ... End(normal) rsp_buf : %s and clear net_dev->rsp_length SSS\n",__FUNCTION__, net_dev->rsp_buffer);

			return 0;
		}
		else if ( timeout > 0)
		{
			//printk("SSS %s ... 3 SSS\n",__FUNCTION__);
			up(&net_dev->cmd_semaphore);

			if (signal_pending(current))
			{
				//printk("SSS %s ... 4 SSS\n",__FUNCTION__);
				return -1;
			}

			if (timeout != MAX_SCHEDULE_TIMEOUT)
				et = jiffies + timeout;

			//interruptible_sleep_on_timeout(&net_dev->cmd_rsp_wait, timeout);
			//wait_event_interruptible(net_dev->cmd_rsp_wait, 0, timeout);
			wait_event_interruptible_timeout(net_dev->cmd_rsp_wait, 0, IXVCOM_CMD_TIMEOUT2);
			//wait_event_interruptible_timeout(net_dev->cmd_rsp_wait, 0, timeout);

			if (timeout != MAX_SCHEDULE_TIMEOUT)
				timeout = et - jiffies;

			//printk("SSS %s ... End(nothing) SSS\n",__FUNCTION__);
		}
		else
		{ // timeout
			up(&net_dev->cmd_semaphore);
			//printk("SSS %s ... End(timeout) SSS\n",__FUNCTION__);
			return -1;
		}
	}
}

static int ixvcom_linectrl(struct ixvcom_hwconf *net_dev, int port, int modem_control)
{
	char rsp_buffer[84];
	int  rsp_length = sizeof(rsp_buffer);

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!net_dev)
		return -EIO;


	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_LINECTRL) < 0)
		return -EIO;

	//VCOM COM RTS/DTR CMD : 05[Port][RTS][DTR]
	if (modem_control & UART_MCR_RTS)
        {
		//active RTS	
		sprintf(net_dev->cmd_buffer, "%s%d%d", net_dev->cmd_buffer, port, 1);		
        }
        else
        {
		//inactive RTS
		sprintf(net_dev->cmd_buffer, "%s%d%d", net_dev->cmd_buffer, port, 0);
        }

	//printk("XXX %s ... WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_LINECTRL) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

	net_dev->cmd_ready = 1;

	if ( waitqueue_active(&net_dev->select_ex_wait))
	{
		wake_up_interruptible( &net_dev->select_ex_wait );
	}

	if (ixvcom_wait_command_completed(net_dev,
					  IXVCOM_VCOM_COMMAND_SET,
					  VCOM_CMD_SET_LINECTRL,
					  IXVCOM_CMD_TIMEOUT,
					  rsp_buffer,
					  &rsp_length))
		return -EIO;

	//printk("XXX %s ... WCC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_LINECTRL) : %s\n",__FUNCTION__, rsp_buffer);

	if ((rsp_buffer[1] != 'O') || (rsp_buffer[2] != 'K') )
		return -EIO;

	return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static void ixvcom_set_termios(struct tty_struct * tty, struct termios * old_termios)
#else
static void ixvcom_set_termios(struct tty_struct * tty, struct ktermios * old_termios)
#endif
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;

	ixvcom_port_init(info, old_termios);
}

static void ixvcom_stop(struct tty_struct * tty)
{
	
}

static void ixvcom_start(struct tty_struct * tty)
{

}

static void ixvcom_hangup(struct tty_struct *tty)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;

	//printk("XXX %s ... start XXX\n",__FUNCTION__);

	if (!info)
		return;

	if (info->flags & ASYNC_CLOSING)
		return;

	info->flags |=  ASYNC_CLOSING;

	ixvcom_shutdown(info);
}

static inline void ixvcom_check_modem_status(struct ixvcom_struct *info, int status)
{
	int is_dcd_changed = 0;

	//printk("XXX %s ... start XXX\n",__FUNCTION__);

	/* update input line counters */
	if ( (info->modem_status & UART_MSR_DSR) != (status & UART_MSR_DSR ))
		info->icount.dsr++;

	if ( (info->modem_status & UART_MSR_DCD) != (status & UART_MSR_DCD ))
	{
		info->icount.dcd++;
		is_dcd_changed = 1;
	}

	if ( (info->modem_status & UART_MSR_CTS) != (status & UART_MSR_CTS ))
		info->icount.cts++;

	info->modem_status = status;
	wake_up_interruptible(&info->delta_msr_wait);

	if ( (info->flags & ASYNC_CHECK_CD) && (is_dcd_changed))
	{
		if ( status & UART_MSR_DCD )
		{
			wake_up_interruptible(&info->open_wait);
		}
		else
		{
			set_bit(IXVCOM_EVENT_HANGUP,&info->event);
			schedule_work(&info->tqueue);
		}
	}
}

static int ixvcom_block_til_ready(struct tty_struct *tty, struct file * filp, struct ixvcom_struct *info)
{
	DECLARE_WAITQUEUE(wait, current);
	int retval;
	int do_clocal = 0;
	struct ixvcom_hwconf *net_dev;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!(net_dev = info->net_dev))
		return -EIO;


	//printk("XXX %s ... 1 XXX\n",__FUNCTION__);
	if ( tty->driver->subtype == SERIAL_TYPE_CALLOUT )
	{
		if ( info->flags & ASYNC_NORMAL_ACTIVE )
			return -EBUSY;

		if ( (info->flags & ASYNC_CALLOUT_ACTIVE) && (info->flags & ASYNC_SESSION_LOCKOUT) && (info->session != MX_SESSION()))
			return -EBUSY;

		if ( (info->flags & ASYNC_CALLOUT_ACTIVE) && (info->flags & ASYNC_PGRP_LOCKOUT) && (info->pgrp != MX_CGRP()) )
			return -EBUSY;

		info->flags |= ASYNC_CALLOUT_ACTIVE;
	//printk("XXX %s ... 2 XXX\n",__FUNCTION__);

		return 0;
	}

	if ( (filp->f_flags & O_NONBLOCK) || (tty->flags & (1 << TTY_IO_ERROR)) )
	{
		if ( info->flags & ASYNC_CALLOUT_ACTIVE )
		{
			return -EBUSY;
		}

		info->flags |= ASYNC_NORMAL_ACTIVE;
		return 0;
	}

	//printk("XXX %s ... 3 XXX\n",__FUNCTION__);
	if ( info->flags & ASYNC_CALLOUT_ACTIVE )
	{
		if ( info->normal_termios.c_cflag & CLOCAL )
			do_clocal = 1;
	}
	else
	{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
                if ( tty->termios.c_cflag & CLOCAL )
#else
                if ( tty->termios->c_cflag & CLOCAL )
#endif
			do_clocal = 1;
	}

	retval = 0;

	add_wait_queue(&info->open_wait, &wait);

	//printk("XXX %s ... 4 XXX\n",__FUNCTION__);
	while(1)
	{
		set_current_state(TASK_INTERRUPTIBLE);

		if ( tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING) )
		{
			if ( !tty_hung_up_p(filp) )
			{
#ifdef SERIAL_DO_RESTART
				if ( info->flags & ASYNC_HUP_NOTIFY )
				{
					retval = -EAGAIN;
				}
				else
					retval = -ERESTARTSYS;
#else
				retval = -EAGAIN;
#endif
			}

			break;
		}

		if ( !(info->flags & ASYNC_CALLOUT_ACTIVE) && !(info->flags & ASYNC_CLOSING) && (do_clocal || (info->modem_status & UART_MSR_DCD)) )
			break;

		if (signal_pending(current))
		{
			retval = -EIO;
			break;
		}

		schedule();
	}

	//printk("XXX %s ... 5 XXX\n",__FUNCTION__);
	set_current_state(TASK_RUNNING);
	remove_wait_queue(&info->open_wait, &wait);

	if (retval)
		return retval;

	info->flags |= ASYNC_NORMAL_ACTIVE;

	//printk("SSS %s ... END SSS\n",__FUNCTION__);
	return 0;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static int ixvcom_port_init(struct ixvcom_struct *info, struct termios *old_termios)
#else
static int ixvcom_port_init(struct ixvcom_struct *info, struct ktermios *old_termios)
#endif
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
	struct termios *termio;
#else
	struct ktermios *termio;
#endif
	int port = 0;
	int32_t baud;
	int modem_status;
	struct ixvcom_hwconf *net_dev;
	char rsp_buffer[84];
	char mode_str[16];
	int rsp_length = sizeof(rsp_buffer);

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	port = info->port;
	net_dev = info->net_dev;

	//printk("XXX %s ... 1 ... Port(%d) init\n", __FUNCTION__, port);

	if ( !info->tty || !net_dev)
	{
		printk("info->tty or net_dev is null\n");
		return -EIO;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
        if (!(termio = &info->tty->termios))
#else
        if (!(termio = info->tty->termios))
#endif
	{
		printk("info->tty->termios is null\n");
		return -EIO;
	}

	/* Get the module's name */
/*
	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_GET_MODULE_NAME) < 0)
        {
                printk("Get VCOM_CMD_GET_MODULE_NAME failed\n");
                return -EIO;
        }

	sprintf(net_dev->cmd_buffer, "%s", net_dev->cmd_buffer);

        net_dev->cmd_ready = 1;
	printk("XXX %s ... 2 WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_DATAFORMAT) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

	if (waitqueue_active(&net_dev->select_ex_wait))
                wake_up_interruptible( &net_dev->select_ex_wait );

        if (ixvcom_wait_command_completed(net_dev,
                                          IXVCOM_VCOM_COMMAND_SET,
                                          VCOM_CMD_SET_DATAFORMAT,
                                          IXVCOM_CMD_TIMEOUT,
                                          rsp_buffer,
                                          &rsp_length))
        {
                printk("wait VCOM_CMD_SET_DATAFORMAT response failed\n");
                return -EIO;
        }

	printk("XXX %s ... 2 WCC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_DATAFORMAT) : %s\n",__FUNCTION__, rsp_buffer);
*/
	/* Set the data format of virtual COM */ 

	if ((termio->c_cflag & CSIZE) == CS5)
	{
		sprintf(mode_str, "%d", 5);
	}
	else if ((termio->c_cflag & CSIZE) == CS6)
	{
		sprintf(mode_str, "%d", 6);
	}
	else if ((termio->c_cflag & CSIZE) == CS7)
	{
		sprintf(mode_str, "%d", 7);
	}
	else if ((termio->c_cflag & CSIZE) == CS8)
	{
		sprintf(mode_str, "%d", 8);
	}

	if (termio->c_cflag & PARENB)
	{
		if (termio->c_cflag & PARODD)
		{
			sprintf(mode_str, "%s%c", mode_str, 'o');
		}
		else
		{
			sprintf(mode_str, "%s%c", mode_str, 'e');
		}
	}
	else
	{
		sprintf(mode_str, "%s%c", mode_str, 'n');
	}

	if (termio->c_cflag & CSTOPB)
	{
		sprintf(mode_str, "%s%d", mode_str, 2);
	}
        else
	{
		sprintf(mode_str, "%s%d", mode_str, 1);
	}


	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_DATAFORMAT) < 0)
        {
                printk("set VCOM_CMD_PORT_INIT failed\n");
                return -EIO;
        }


	//VCOM Data Format CMD : 03[Port][Data Bits][Data Parity][Stop Bit]
	sprintf(net_dev->cmd_buffer, "%s%d%s", net_dev->cmd_buffer, port, mode_str);
	
	net_dev->cmd_ready = 1;

	//printk("XXX %s ... 2 WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_DATAFORMAT) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

	if (waitqueue_active(&net_dev->select_ex_wait))
                wake_up_interruptible( &net_dev->select_ex_wait );

        if (ixvcom_wait_command_completed(net_dev,
                                          IXVCOM_VCOM_COMMAND_SET,
                                          VCOM_CMD_SET_DATAFORMAT,
                                          IXVCOM_CMD_TIMEOUT,
                                          rsp_buffer,
                                          &rsp_length))
        {
                printk("wait VCOM_CMD_SET_DATAFORMAT response failed\n");
                return -EIO;
        }

	//printk("XXX %s ... 2 WCC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_DATAFORMAT) : %s\n",__FUNCTION__, rsp_buffer);

	if ((rsp_buffer[1] != 'O') || (rsp_buffer[2] != 'K') )
                return -EIO;

	/* Set the baud rate of virtual COM */ 

	switch ( termio->c_cflag & (CBAUD|CBAUDEX))
	{
		case B921600:
			baud = 921600L;
			break;
		case B460800:
			baud = 460800;
			break;
		case B230400:
			baud = 230400L;
			break;
		case B115200:
			baud = 115200L;
			break;
		case B57600:
			baud = 57600L;
			break;
		case B38400:
			baud = 38400L;

			if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI )
			{
				baud = 57600L;
			}

			if ( (info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI )
			{
				baud = 115200L;
			}

#ifdef ASYNC_SPD_SHI
			if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI)
			{
				baud = 230400L;
			}
#endif

#ifdef ASYNC_SPD_WARP
			if ((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP)
			{
				baud = 460800L;
			}
#endif
		        break;

		case B19200:
		        baud = 19200L;
		        break;

		case B9600:
			baud = 9600L;
			break;

		case B4800:
			baud = 4800L;
			break;

		case B2400:
			baud = 2400L;
			break;

		case B1800:
			baud = 1800L;
			break;

		case B1200:
			baud = 1200L;
			break;

		case B600:
			baud = 600L;
			break;

		case B300:
			baud = 300L;
			break;

		case B200:
			baud = 200L;
			break;

		case B150:
			baud = 150L;
			break;

		case B134:
			baud = 134L;
			break;

		case B110:
			baud = 110L;
			break;

		case B75:
			baud = 75L;
			break;
		
		case B50:
			baud = 50L;
			break;

		default:
			baud = 0;
	}

	if (baud > 921600L)
	{
		//printk("XXX %s ... Run Here\n",__FUNCTION__);
		termio->c_cflag &= ~(CBAUD|CBAUDEX);
		termio->c_cflag |= old_termios->c_cflag &(CBAUD|CBAUDEX);
	}


	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BAUD) < 0)
	{
		printk("set VCOM_CMD_SET_BAURD failed\n");
		return -EIO;
	}

	//printk("XXX %s ... 2 BAUD : %d\n",__FUNCTION__, baud);
	sprintf(net_dev->cmd_buffer, "%s%d%d", net_dev->cmd_buffer, port, baud);

	//printk("XXX %s ... 3 WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BAUD) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

	net_dev->cmd_ready = 1;

        if (waitqueue_active(&net_dev->select_ex_wait))
                wake_up_interruptible( &net_dev->select_ex_wait );

        if (ixvcom_wait_command_completed(net_dev,
                                          IXVCOM_VCOM_COMMAND_SET,
                                          VCOM_CMD_SET_BAUD,
                                          IXVCOM_CMD_TIMEOUT,
                                          rsp_buffer,
                                          &rsp_length))
        {
                printk("wait VCOM_CMD_SET_BAUD response failed\n");
                return -EIO;
        }
	
	//printk("XXX %s ... 3 WCC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BAUD) : %s\n",__FUNCTION__, rsp_buffer);

	if ((rsp_buffer[1] != 'O') || (rsp_buffer[2] != 'K') )
                return -EIO;

	/* Set LineControl(RTS/DTR) */
/*
	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_LINECTRL) < 0)
                return -EIO;

	//VCOM COM RTS/DTR CMD : 05[Port][RTS][DTR]
	if (info->modem_control & UART_MCR_RTS)
	{
		//active RTS
		sprintf(net_dev->cmd_buffer, "%s%d%d", net_dev->cmd_buffer, port, 1);      
	}
	else
	{
		//inactive RTS
		sprintf(net_dev->cmd_buffer, "%s%d%d", net_dev->cmd_buffer, port, 0);
	}

	printk("XXX %s ... 4 WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_LINECTRL) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

	net_dev->cmd_ready = 1;

	if (waitqueue_active(&net_dev->select_ex_wait))
		wake_up_interruptible( &net_dev->select_ex_wait );

	if (ixvcom_wait_command_completed(net_dev,
					  IXVCOM_VCOM_COMMAND_SET,
					  VCOM_CMD_SET_LINECTRL,
					  IXVCOM_CMD_TIMEOUT,
					  rsp_buffer,
					  &rsp_length))
	{
		printk("wait VCOM_CMD_SET_LINECTRL response failed\n");
		return -EIO;
	}

	printk("XXX %s ... 4 WCC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_LINECTRL) : %s\n",__FUNCTION__, rsp_buffer);

	if ((rsp_buffer[1] != 'O') || (rsp_buffer[2] != 'K') )
                return -EIO;

*/
	/* Set CTS/RTS follow control */
/*
        if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_FOLLOWCTRL) < 0)
                return -EIO;

	//VCOM CTS/RTS follow control CMD : 28[Port][CTS][RTS]
	if (termio->c_cflag & CRTSCTS)
        {
		//Enable CTS/RTS
		sprintf(net_dev->cmd_buffer, "%s%d%d%d", net_dev->cmd_buffer, port, 1, 1); 
        }
        else
        {
		//Disable CTS/RTS
		sprintf(net_dev->cmd_buffer, "%s%d%d%d", net_dev->cmd_buffer, port, 0, 0); 
        }

	printk("XXX %s ... 5 WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_FOLLOWCTRL) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

        net_dev->cmd_ready = 1;

        if (waitqueue_active(&net_dev->select_ex_wait))
                wake_up_interruptible( &net_dev->select_ex_wait );

        if (ixvcom_wait_command_completed(net_dev,
                                          IXVCOM_VCOM_COMMAND_SET,
                                          VCOM_CMD_SET_FOLLOWCTRL,
                                          IXVCOM_CMD_TIMEOUT,
                                          rsp_buffer,
                                          &rsp_length))
        {
                printk("wait VCOM_CMD_SET_FOLLOWCTRL response failed\n");
                return -EIO;
        }

        printk("XXX %s ... 5 WCC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_FOLLOWCTRL) : %s\n",__FUNCTION__, rsp_buffer);

        if ((rsp_buffer[1] != 'O') || (rsp_buffer[2] != 'K') )
                return -EIO;
*/

	modem_status = 0;

	ixvcom_check_modem_status(info, modem_status);

	if ( termio->c_cflag & CLOCAL )
	{
		info->flags &= ~ASYNC_CHECK_CD;
	}
	else
	{
		info->flags |= ASYNC_CHECK_CD;
	}

	if ( !info->tty)
	{
		printk("info->tty is null\n");
		return -EIO;
	}

//	printk("SSS %s ... c_cflag : 0x%x END SSS\n",__FUNCTION__, termio->c_cflag);
	return 0;
}

static int ixvcom_port_shutdown(struct ixvcom_struct *info)
{
#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
	struct termios *termio;
#else
	struct ktermios *termio;
#endif
	struct ixvcom_hwconf *net_dev;
	char rsp_buffer[84];
	int  rsp_length = sizeof(rsp_buffer);

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	net_dev = info->net_dev;

	if ( !info->tty || !net_dev)
	{
		return -EIO;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
        if (!(termio = &info->tty->termios))
#else
        if (!(termio = info->tty->termios))
#endif
	{
		return -EIO;
	}

	ixvcom_wait_and_set_command(net_dev,
                                    IXVCOM_LOCAL_COMMAND_SET,
                                    LOCAL_CMD_TTY_UNUSED);

	net_dev->cmd_ready = 1;

	if ( waitqueue_active(&net_dev->select_ex_wait))
		wake_up_interruptible( &net_dev->select_ex_wait );

	ixvcom_wait_command_completed(net_dev,
				      IXVCOM_LOCAL_COMMAND_SET,
				      LOCAL_CMD_TTY_UNUSED,
				      IXVCOM_CMD_TIMEOUT,
				      rsp_buffer,
				      &rsp_length);

	net_dev->flag &= ~IXVCOM_NET_TTY_INUSED;

	return 0;
}


static int ixvcom_startup(struct ixvcom_struct * info, struct file *filp, struct tty_struct *tty)
{
	unsigned long page;
	struct ixvcom_hwconf *net_dev;
	char rsp_buffer[84];
	int  rsp_length = sizeof(rsp_buffer);
	int cnt = 0;


	DECLARE_WAITQUEUE(wait, current);
	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!(net_dev = info->net_dev))
	{
		printk("info->net_dev is null\n");
		return -EIO;
	}

	add_wait_queue(&net_dev->initialize_wait, &wait);

	while (test_and_set_bit(IXVCOM_NET_DO_INITIALIZE, &net_dev->flag))
	{
		set_current_state(TASK_INTERRUPTIBLE);

		if (signal_pending(current))
		{
			printk("signal pending break\n");
			break;
		}
		
		schedule();
	}

	set_current_state(TASK_RUNNING);
	remove_wait_queue(&net_dev->initialize_wait, &wait);

	info->tty = tty;

	//printk("SSS %s ... 1 SSS\n",__FUNCTION__);
	if (signal_pending(current))
	{
		printk("signal pending occurred\n");

		if ( waitqueue_active(&net_dev->initialize_wait))
			wake_up_interruptible( &net_dev->initialize_wait );

		return -EIO;
	}

	if ( info->flags & ASYNC_INITIALIZED )
	{
		clear_bit(IXVCOM_NET_DO_INITIALIZE, &net_dev->flag);

		if ( waitqueue_active(&net_dev->initialize_wait))
			wake_up_interruptible( &net_dev->initialize_wait );

		return 0;
	}

	page = __get_free_page(GFP_KERNEL);

	if (!page)
	{
		clear_bit(IXVCOM_NET_DO_INITIALIZE, &net_dev->flag);

		if ( waitqueue_active(&net_dev->initialize_wait))
			wake_up_interruptible( &net_dev->initialize_wait );

		printk("page get free failed\n");

		return -ENOMEM;
	}

	//printk("SSS %s ... 2 SSS\n",__FUNCTION__);
	if (!(net_dev->flag & IXVCOM_NET_TTY_INUSED))
	{
		//printk("XXX %s ... 1 XXX\n",__FUNCTION__);
		net_dev->cmd_buffer[0] = HexToAscii(0);
		//use "LOCAL_CMD_TTY_USED" to open userspace TCP socket port 10000, 1000X
		//printk("SSS %s ... 2-1 SSS\n",__FUNCTION__);
		ixvcom_wait_and_set_command(net_dev, 
					    IXVCOM_LOCAL_COMMAND_SET,
					    LOCAL_CMD_TTY_USED);

		//net_dev->cmd_buffer[2] = 0;
		net_dev->cmd_ready = 1;

		if (waitqueue_active(&net_dev->select_ex_wait))
		{
			wake_up_interruptible( &net_dev->select_ex_wait );
		}

		//printk("SSS %s ... 2-2 SSS\n",__FUNCTION__);
		if (ixvcom_wait_command_completed(net_dev,
						  IXVCOM_LOCAL_COMMAND_SET,
						  LOCAL_CMD_TTY_USED,
						  IXVCOM_CMD_TIMEOUT,
						  rsp_buffer,
						  &rsp_length) != 0)
		{
			printk("wait for LOCAL_CMD_TTY_USED response failed\n");
			ixvcom_wait_and_set_command(net_dev,
						    IXVCOM_LOCAL_COMMAND_SET,
						    LOCAL_CMD_TTY_UNUSED);

			net_dev->cmd_buffer[2] = 0;
			net_dev->cmd_ready = 1;

			if ( waitqueue_active(&net_dev->select_ex_wait))
			{
				wake_up_interruptible( &net_dev->select_ex_wait );
			}

			ixvcom_wait_command_completed(net_dev,
						      IXVCOM_LOCAL_COMMAND_SET,
						      LOCAL_CMD_TTY_UNUSED,
						      IXVCOM_CMD_TIMEOUT,
						      rsp_buffer,
						      &rsp_length);

			goto startup_err;
		}

		//printk("XXX %s ... 2 XXX\n",__FUNCTION__);
		//printk("SSS %s ... 2-3 SSS\n",__FUNCTION__);
		net_dev->flag |= IXVCOM_NET_TTY_INUSED;
	}
	else
	{
		//printk("XXX %s ... 3 XXX\n",__FUNCTION__);
		while ((net_dev->cmd_ready == 1) && (cnt++ < 10))
		{
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ/100);
		}
	}

	net_dev->flag &= ~IXVCOM_NET_DO_SESSION_RECOVERY;
	info->modem_status = 0;
	info->modem_control = 0;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
        if (info->tty->termios.c_cflag & CBAUD)
#else
        if (info->tty->termios->c_cflag & CBAUD)
#endif
		info->modem_control = UART_MCR_DTR | UART_MCR_RTS;



	//printk("SSS %s ... 3 SSS\n",__FUNCTION__);
/*
	if (ixvcom_port_init(info, 0) != 0)
	{
		printk("ixvcom port init failed\n");
		goto startup_err;
	}
*/

	//printk("XXX %s ... 4 XXX\n",__FUNCTION__);
	if (info->uart_type == PORT_16550A)
	{
/*
		if (net_dev->server_type == CN2500)
			info->xmit_fifo_size = 64;
		else
*/
			info->xmit_fifo_size = 16;
	}
	else
		info->xmit_fifo_size = 1;

/*
	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_TX_FIFO) < 0)
	{
		printk("Set VCOM_CMD_TX_FIFO failed\n");
		goto startup_err;
	}

	net_dev->cmd_buffer[2] = 1;
	net_dev->cmd_buffer[3] = info->xmit_fifo_size;
	net_dev->cmd_ready = 1;

	if ( waitqueue_active(&net_dev->select_ex_wait))
	{
		wake_up_interruptible( &net_dev->select_ex_wait );
	}

	rsp_length = sizeof(rsp_buffer);

	if (ixvcom_wait_command_completed(net_dev,
					  IXVCOM_VCOM_COMMAND_SET,
					  VCOM_CMD_TX_FIFO,
					  IXVCOM_CMD_TIMEOUT,
					  rsp_buffer,
					  &rsp_length) != 0)
	{
		printk("Wait for VCOM_CMD_TX_FIFO response failed\n");
		goto startup_err;
	}
*/

	if ( info->xmit_buf )
		free_page(page);
	else
		info->xmit_buf = (unsigned char *)page;


	if ( info->tty )
        	test_and_clear_bit(TTY_IO_ERROR, &info->tty->flags);
    
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	info->flags |= ASYNC_INITIALIZED;
	clear_bit(IXVCOM_NET_DO_INITIALIZE, &net_dev->flag);

	if ( waitqueue_active(&net_dev->initialize_wait))
		wake_up_interruptible( &net_dev->initialize_wait );

	//printk("SSS %s ... END SSS\n",__FUNCTION__);

	return 0;

startup_err:
	free_page(page);

	clear_bit(IXVCOM_NET_DO_INITIALIZE, &net_dev->flag);

	if ( waitqueue_active(&net_dev->initialize_wait))
		wake_up_interruptible( &net_dev->initialize_wait );

	return -EIO;
}

static void ixvcom_shutdown(struct ixvcom_struct * info)
{
	struct ixvcom_hwconf *net_dev = info->net_dev;
	unsigned long flags;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!net_dev)
		printk("net_dev is null\n");

	while (test_and_set_bit(IXVCOM_NET_DO_INITIALIZE, &net_dev->flag))
	{
		if (signal_pending(current))
			break;

		set_current_state(TASK_INTERRUPTIBLE);
		schedule_timeout(HZ/100);
	}


	if ( !(info->flags & ASYNC_INITIALIZED) )
	{
		goto shutdown_ok;
	}

	spin_lock_irqsave(&info->tx_lock, flags);

	if ( info->xmit_buf )
	{
		free_page((unsigned long)info->xmit_buf);
		info->xmit_buf = 0;
	}

	spin_unlock_irqrestore(&info->tx_lock, flags);

	if ( info->tty )
	{
		set_bit(TTY_IO_ERROR, &info->tty->flags);
		ixvcom_unthrottle(info->tty);
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
        if (!info->tty || (info->tty->termios.c_cflag & HUPCL))
#else
        if (!info->tty || (info->tty->termios->c_cflag & HUPCL))
#endif
	{
		info->modem_control &= ~(UART_MCR_DTR | UART_MCR_RTS);
	}

	ixvcom_port_shutdown(info);

shutdown_ok:
    ;

	info->flags &= ~(ASYNC_NORMAL_ACTIVE|ASYNC_CALLOUT_ACTIVE|ASYNC_INITIALIZED|ASYNC_CLOSING);

	down (&info->rx_semaphore);
	info->tty = 0;
	up (&info->rx_semaphore);

	clear_bit(IXVCOM_NET_DO_INITIALIZE,&net_dev->flag);
	wake_up_interruptible(&info->open_wait);
	wake_up_interruptible(&info->close_wait);

	if ( waitqueue_active(&net_dev->initialize_wait))
		wake_up_interruptible( &net_dev->initialize_wait );
	
	//printk("SSS %s ... END SSS\n",__FUNCTION__);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
static int ixvcom_break(struct tty_struct *ttyinfo, int break_state)
#else
static void ixvcom_break(struct tty_struct *ttyinfo, int break_state)
#endif
{
	struct ixvcom_struct *info;
	struct ixvcom_hwconf *net_dev;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if ( !ttyinfo  )
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
		return (-EFAULT);
#else
		return;
#endif

	info = (struct ixvcom_struct *)ttyinfo->driver_data;

	if (!info)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
		return (-EFAULT);
#else
		return;
#endif


	if (!(net_dev = info->net_dev))
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
		return (-EFAULT);
#else
		return;
#endif


	if (break_state == -1)
	{
		ixvcom_start_break(net_dev, info->port);
	}
	else
	{
		ixvcom_stop_break(net_dev, info->port);
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
	return (0);
#endif
}

static void ixvcom_throttle(struct tty_struct * tty)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;
	struct ixvcom_hwconf *net_dev;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
		return;

	net_dev = info->net_dev;

	if (!net_dev)
		return;

	net_dev->rx_ready = 0;
}

static void ixvcom_unthrottle(struct tty_struct * tty)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;
	struct ixvcom_hwconf *net_dev;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
		return;

	net_dev = info->net_dev;

	if (!net_dev)
		return;

	net_dev->rx_ready = 1;

	if ( waitqueue_active(&net_dev->select_out_wait))
		wake_up_interruptible( &net_dev->select_out_wait );
}

static long ixvcom_wait_oqueue(struct ixvcom_struct *info, long timeout)
{
	struct ixvcom_hwconf *net_dev;
	//long et = 0;
	//uint32_t    tout;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!(net_dev = info->net_dev))
		return -EIO;

/*
	if (ixvcom_wait_and_set_command(net_dev,
					IXVCOM_VCOM_COMMAND_SET,
					VCOM_CMD_WAIT_OQUEUE) < 0)
		return -EIO;

	if (timeout != MAX_SCHEDULE_TIMEOUT)
	{
		if (timeout < HZ/10)  // at least wait for 100 ms
			timeout = HZ/10;

		et = jiffies + timeout;
	}

	if (timeout != MAX_SCHEDULE_TIMEOUT)
		tout = (uint32_t)timeout;
	else
		tout = 0x7FFFFFFF;

	net_dev->cmd_buffer[2] = 4;
	memcpy(&net_dev->cmd_buffer[3], (void *)&tout, 4);
*/
	net_dev->wait_oqueue_responsed = 0;
/*
	net_dev->cmd_ready = 1;

	if (waitqueue_active(&net_dev->select_ex_wait))
		wake_up_interruptible( &net_dev->select_ex_wait );

	while (net_dev->cmd_ready == 1)
	{
        	//interruptible_sleep_on_timeout(&net_dev->cmd_rsp_wait, 1);
		wait_event_interruptible_timeout(net_dev->cmd_rsp_wait, 0, 1);

		if (timeout != MAX_SCHEDULE_TIMEOUT)
		{
			if (jiffies > et)
				return -EIO;
		}

	}

	if (timeout != MAX_SCHEDULE_TIMEOUT)
		timeout += 10;

	net_dev->cmd_buffer[0] = 0;
*/
/*
	do
	{
		if (net_dev->wait_oqueue_responsed == 0)
        	{
			//timeout = interruptible_sleep_on_timeout(&net_dev->cmd_rsp_wait, timeout);
			timeout = wait_event_interruptible_timeout(net_dev->cmd_rsp_wait, 0, timeout);
	
			if (net_dev->wait_oqueue_responsed)
			{
				return (net_dev->oqueue);
			}
		}
	        else
		{
			return (net_dev->oqueue);
		}
	}

	while (timeout > 0);
	
	return -EIO;
*/
	return 0;

}

static void ixvcom_start_break(struct ixvcom_hwconf *net_dev, int port)
{
	char rsp_buffer[84];
	int rsp_length = sizeof (rsp_buffer);

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	/* Send break to COM */
	ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BREAK);

	//VCOM Enable Break CMD : 27[Port][Enable/Disable Break]
	sprintf(net_dev->cmd_buffer, "%s%d%d", net_dev->cmd_buffer, port, 1);	

	//printk("XXX %s ... WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BREAK) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

	net_dev->cmd_ready = 1;

	if ( waitqueue_active(&net_dev->select_ex_wait))
	{
		wake_up_interruptible( &net_dev->select_ex_wait );
	}

	if (ixvcom_wait_command_completed(net_dev,
					  IXVCOM_VCOM_COMMAND_SET,
					  VCOM_CMD_SET_BREAK,
					  IXVCOM_CMD_TIMEOUT,
					  rsp_buffer,
					  &rsp_length))
		return;

	//printk("XXX %s ... WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BREAK) : %s\n",__FUNCTION__, rsp_buffer);

	if ((rsp_buffer[1] != 'O') || (rsp_buffer[2] != 'K') )
		return;

	return;
}

static void ixvcom_stop_break(struct ixvcom_hwconf *net_dev, int port)
{
	char rsp_buffer[84];
	int rsp_length = sizeof (rsp_buffer);

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BREAK);

        //VCOM Disable Break CMD : 27[Port][Enable/Disable Break]
	sprintf(net_dev->cmd_buffer, "%s%d%d", net_dev->cmd_buffer, port, 0);
        
	//printk("XXX %s ... WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BREAK) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

	net_dev->cmd_ready = 1;

	if ( waitqueue_active(&net_dev->select_ex_wait))
		wake_up_interruptible( &net_dev->select_ex_wait );

	rsp_length = sizeof(rsp_buffer);

	if (ixvcom_wait_command_completed(net_dev,
					  IXVCOM_VCOM_COMMAND_SET,
					  VCOM_CMD_SET_BREAK,
					  IXVCOM_CMD_TIMEOUT,
					  rsp_buffer,
					  &rsp_length))
		return;

	//printk("XXX %s ... WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_SET_BREAK) : %s\n",__FUNCTION__, rsp_buffer);

	if ((rsp_buffer[1] != 'O') || (rsp_buffer[2] != 'K') )
		return;

	return;
}


static void ixvcom_send_break(struct ixvcom_struct * info, int duration)
{
	int port = 0;
	struct ixvcom_hwconf *net_dev;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!(net_dev = info->net_dev))
		return;

	port = info->port;

	ixvcom_start_break(net_dev, port);

	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(duration);

	ixvcom_stop_break(net_dev, port);
}

static int ixvcom_get_serial_info(struct ixvcom_struct * info, struct serial_struct * retinfo)
{
	struct serial_struct tmp;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);
	
	if ( !retinfo )
		return -EFAULT;

	memset(&tmp, 0, sizeof(tmp));
	tmp.type = info->uart_type;
	tmp.line = info->port;
	tmp.flags = info->flags;
	tmp.close_delay = info->close_delay;
	tmp.closing_wait = info->closing_wait;
	tmp.custom_divisor = info->custom_divisor;
	tmp.baud_base = info->baud_base;
	tmp.hub6 = 0;

	if (copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
		return -EFAULT;
	
	return 0;
}

static int ixvcom_set_serial_info(struct ixvcom_struct *info, struct serial_struct *new_info)
{
	struct serial_struct new_serial;
	unsigned int flags;
	int retval = 0;
	//char rsp_buffer[84];
	//int rsp_length;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);


	if ( !new_info)
		return -EFAULT;

	if (copy_from_user(&new_serial, new_info, sizeof(new_serial)))
		return -EFAULT;

	flags = info->flags & ASYNC_SPD_MASK;

	if ( !capable(CAP_SYS_ADMIN))
	{
		if ((new_serial.close_delay != info->close_delay) || ((new_serial.flags & ~ASYNC_USR_MASK) != (info->flags & ~ASYNC_USR_MASK)) )
		return -EPERM;

		info->flags = ((info->flags & ~ASYNC_USR_MASK) | (new_serial.flags & ASYNC_USR_MASK));
	}
	else
	{
		info->flags = ((info->flags & ~ASYNC_FLAGS) | (new_serial.flags & ASYNC_FLAGS));
		info->close_delay = new_serial.close_delay * HZ/100;

		if (new_serial.closing_wait == ASYNC_CLOSING_WAIT_NONE)
			info->closing_wait = ASYNC_CLOSING_WAIT_NONE;
		else
			info->closing_wait = new_serial.closing_wait * HZ/100;
	}

	info->uart_type = new_serial.type;
	
	if (info->uart_type == PORT_16550A)
	{
/*
		if (info->net_dev)
		{
				info->xmit_fifo_size = 16;
		}
		else
*/
			info->xmit_fifo_size = 16;
	}
	else
	{
		info->xmit_fifo_size = 1;
	}	

	if ( info->flags & ASYNC_INITIALIZED )
	{
		if ( flags != (info->flags & ASYNC_SPD_MASK) )
		{
			retval = ixvcom_port_init(info,0);
		}

		if (info->net_dev)
		{
/*
			ixvcom_wait_and_set_command(info->net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_TX_FIFO);

			info->net_dev->cmd_buffer[2] = 1;
			info->net_dev->cmd_buffer[3] = info->xmit_fifo_size;
			info->net_dev->cmd_ready = 1;

			if ( waitqueue_active(&info->net_dev->select_ex_wait))
				wake_up_interruptible(&info->net_dev->select_ex_wait);

			ixvcom_wait_command_completed(info->net_dev,
						      IXVCOM_VCOM_COMMAND_SET,
						      VCOM_CMD_TX_FIFO,
						      IXVCOM_CMD_TIMEOUT,
						      rsp_buffer,
						      &rsp_length);
*/
		}
	}

	info->custom_divisor = new_serial.custom_divisor;

	if (info->custom_divisor == 0)
		info->baud_base = 921600L;
	else
		info->baud_base = new_serial.baud_base;

	//printk("SSS %s ... END SSS\n",__FUNCTION__);
	return retval;
}

static int ixvcom_get_lsr_info(struct ixvcom_struct *info, unsigned int *value)
{
	unsigned int result = 0;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (ixvcom_wait_oqueue(info, 0) == 0)
		result = TIOCSER_TEMT;

	put_user(result, value);

	return 0;
}
/*
static int ixvcom_setxon_xoff(struct ixvcom_struct *info, int cmd)
{

	char rsp_buffer[84];
	int  rsp_length = sizeof(rsp_buffer);
	struct ixvcom_hwconf *net_dev;

	printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!(net_dev = info->net_dev))
		return -EIO;

	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET,cmd) < 0)
		return -EIO;

	net_dev->cmd_buffer[2] = 0;
	net_dev->cmd_ready = 1;

	if ( waitqueue_active(&net_dev->select_ex_wait))
	{
		wake_up_interruptible( &net_dev->select_ex_wait );
	}

	if (ixvcom_wait_command_completed(net_dev,
					  IXVCOM_VCOM_COMMAND_SET,
					  cmd,
					  IXVCOM_CMD_TIMEOUT,
					  rsp_buffer,
					  &rsp_length))
		return -EIO;

	if (rsp_length != 4)
		return -EIO;

	if ((rsp_buffer[2] != 'O') || (rsp_buffer[3] != 'K') )
		return -EIO;

	return 0;
}
*/

static void ixvcom_flush_buffer(struct tty_struct * tty)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;
	struct ixvcom_hwconf *net_dev;
	char rsp_buffer[84];
	int rsp_length = sizeof(rsp_buffer);
	unsigned long flags;
	int port;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
		return;

	port = info->port;

	spin_lock_irqsave(&info->tx_lock, flags);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	spin_unlock_irqrestore(&info->tx_lock, flags);

	wake_up_interruptible(&tty->write_wait);

	if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
	tty->ldisc->ops->write_wakeup )
		(tty->ldisc->ops->write_wakeup)(tty);
#else
	tty->ldisc.write_wakeup )
		(tty->ldisc.write_wakeup)(tty);
#endif

	if (!(net_dev = info->net_dev))
		return;

	net_dev->tx_ready = 0;

	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_FLUSH_TXRXBUFFER) < 0)
		return;

	//VCOM Clear TX/RX Buffer CMD : 35[Port][TX][RX]
	sprintf(net_dev->cmd_buffer, "%s%d%d%d", net_dev->cmd_buffer, port, 1, 1);

	//printk("XXX %s ... WASC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_FLUSH_TXRXBUFFER) : %s\n",__FUNCTION__, net_dev->cmd_buffer);

	net_dev->cmd_ready = 1;

	if ( waitqueue_active(&net_dev->select_ex_wait))
	{
		wake_up_interruptible( &net_dev->select_ex_wait );
	}

	ixvcom_wait_command_completed(net_dev,
				      IXVCOM_VCOM_COMMAND_SET,
				      VCOM_CMD_FLUSH_TXRXBUFFER,
				      IXVCOM_CMD_TIMEOUT,
				      rsp_buffer,
				      &rsp_length);
	
	//printk("XXX %s ... WCC(IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_FLUSH_TXRXBUFFER) : %s\n",__FUNCTION__, rsp_buffer);

	//printk("SSS %s ... END SSS\n",__FUNCTION__);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39))
static int ixvcom_tiocmget(struct tty_struct *tty)
#else
static int ixvcom_tiocmget(struct tty_struct *tty, struct file *file)
#endif
{
	struct ixvcom_struct *info = (struct ixvcom_struct *) tty->driver_data;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
	{
		printk("info is null\n");
		return -EINVAL;
	}

	if (tty->index == VXCOM_MAX_PORTS)
		return -ENOIOCTLCMD;

	if (tty->flags & (1 << TTY_IO_ERROR))
		return -EIO;

	return ((info->modem_control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
		((info->modem_control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
		((info->modem_status  & UART_MSR_DCD) ? TIOCM_CAR : 0) |
		((info->modem_status  & UART_MSR_RI)  ? TIOCM_RNG : 0) |
		((info->modem_status  & UART_MSR_DSR) ? TIOCM_DSR : 0) |
		((info->modem_status  & UART_MSR_CTS) ? TIOCM_CTS : 0);
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39))
static int ixvcom_tiocmset(struct tty_struct *tty, unsigned int set, unsigned int clear)
#else
static int ixvcom_tiocmset(struct tty_struct *tty, struct file *file, unsigned int set, unsigned int clear)
#endif
{
	int port = 0;
	struct ixvcom_struct *info = (struct ixvcom_struct *) tty->driver_data;
	struct ixvcom_hwconf *net_dev;


	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
	{
		printk("info is null\n");
		return (-EINVAL);
	}

	port = info->port;

	if (!(net_dev = info->net_dev))
		return -EINVAL;

	if (tty->index == VXCOM_MAX_PORTS)
		return -ENOIOCTLCMD;

	if (tty->flags & (1 << TTY_IO_ERROR))
		return -EIO;

	if (set & TIOCM_RTS)
		info->modem_control |= UART_MCR_RTS;

	if (set & TIOCM_DTR)
		info->modem_control |= UART_MCR_DTR;

	if (clear & TIOCM_RTS)
		info->modem_control &= ~UART_MCR_RTS;

	if (clear & TIOCM_DTR)
		info->modem_control &= ~UART_MCR_DTR;

	return ixvcom_linectrl(net_dev, port, info->modem_control);
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26))
static void ixvcom_put_char(struct tty_struct * tty, unsigned char ch)
#else
static int ixvcom_put_char(struct tty_struct * tty, unsigned char ch)
#endif
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;
	struct ixvcom_hwconf *net_dev;
	unsigned long        flags;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
        return 0;
#else
        return;
#endif

	if ( !tty || !info->xmit_buf )
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
	return 0;
#else
	return;
#endif

	spin_lock_irqsave(&info->tx_lock, flags);

	if (!info->xmit_buf)
	{
		spin_unlock_irqrestore(&info->tx_lock, flags);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
		return 0;
#else
		return;
#endif
	}

	if ( info->xmit_cnt >= SERIAL_XMIT_SIZE - 1 )
	{
		spin_unlock_irqrestore(&info->tx_lock, flags);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
		return 0;
#else
		return;
#endif
	}

	net_dev = info->net_dev;

	if (!net_dev)
	{
		spin_unlock_irqrestore(&info->tx_lock, flags);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
		return 0;
#else
		return;
#endif
	}

	info->xmit_buf[info->xmit_head++] = ch;
	info->xmit_head &= SERIAL_XMIT_SIZE - 1;
	info->xmit_cnt++;

	net_dev->tx_ready = 1;

	if ( waitqueue_active(&net_dev->select_in_wait))
		wake_up_interruptible( &net_dev->select_in_wait );

	spin_unlock_irqrestore(&info->tx_lock, flags);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,26))
	return 1;
#endif
}

static void ixvcom_wait_until_sent(struct tty_struct * tty, int timeout)
{
	struct ixvcom_struct *info;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if ((info = (struct ixvcom_struct *)tty->driver_data))
		ixvcom_wait_oqueue(info,timeout);
}

static int ixvcom_write_room(struct tty_struct * tty)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;
	int ret;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
		return 0;

	ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;

	if ( ret < 0 )
		ret = 0;

	return(ret);
}



static int ixvcom_chars_in_buffer(struct tty_struct * tty)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;

	//printk("SSS %s ... STATR SSS\n",__FUNCTION__);

	if (!info)
		return -EIO;

	return(info->xmit_cnt);
}

static void ixvcom_flush_chars(struct tty_struct * tty)
{
	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	return;
}

static void ixvcom_ldisc_flush_buffer(struct tty_struct * tty)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;
	struct ixvcom_hwconf *net_dev;
	unsigned long flags;
	int port;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
		return;

	port = info->port;

	spin_lock_irqsave(&info->tx_lock, flags);
	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
 
	spin_unlock_irqrestore(&info->tx_lock, flags);

	wake_up_interruptible(&tty->write_wait);

	if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
		tty->ldisc->ops->write_wakeup )
			(tty->ldisc->ops->write_wakeup)(tty);
#else
		tty->ldisc.write_wakeup )
			(tty->ldisc.write_wakeup)(tty);
#endif

	if (!(net_dev=info->net_dev))
		return;

	net_dev->tx_ready = 0;

	if (net_dev->flag & IXVCOM_NET_DO_SESSION_RECOVERY)
		return;

	if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, VCOM_CMD_FLUSH_TXRXBUFFER) < 0)
                return;

        //VCOM Clear TX/RX Buffer CMD : 35[Port][TX][RX]
	sprintf(net_dev->cmd_buffer, "%s%d%d%d", net_dev->cmd_buffer, port, 1, 1);

	net_dev->cmd_ready = 1;

	if ( waitqueue_active(&net_dev->select_ex_wait))
	{
		wake_up_interruptible( &net_dev->select_ex_wait );
	}

	set_current_state(TASK_INTERRUPTIBLE);
	schedule_timeout(HZ/100);
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10))
static int ixvcom_write(struct tty_struct * tty, int from_user, const unsigned char * buf, int count)
#else
static int ixvcom_write(struct tty_struct * tty, const unsigned char * buf, int count)
#endif
{
	int c, total = 0, ret;
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;
	struct ixvcom_hwconf *net_dev;
	unsigned long flags;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10))
	int from_user = 0;
#endif

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
		return 0;

	if ( !tty || !info->xmit_buf || !ixvcom_tmp_buf )
		return(0);

	net_dev = info->net_dev;

	if (!net_dev)
		return 0;

	if ( from_user )
		down(&ixvcom_tmp_buf_sem);

	while (1)
	{
		c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1, SERIAL_XMIT_SIZE - info->xmit_head));

		if ( c <= 0 )
			break;

		if (from_user)
		{
			ret = copy_from_user(ixvcom_tmp_buf, buf, c);
			memcpy(info->xmit_buf + info->xmit_head, ixvcom_tmp_buf, c);
		}
		else
		{
			memcpy(info->xmit_buf + info->xmit_head, buf, c);
		}

		spin_lock_irqsave(&info->tx_lock, flags);
		info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE - 1);
		info->xmit_cnt += c;
		spin_unlock_irqrestore(&info->tx_lock, flags);

		buf += c;
		count -= c;
		total += c;
	}

	if (info->xmit_cnt )
	{
		net_dev->tx_ready = 1;
		
		if (waitqueue_active(&net_dev->select_in_wait))
			wake_up_interruptible( &net_dev->select_in_wait );
	}

	if (from_user)
		up(&ixvcom_tmp_buf_sem);

	return total;
}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,39))
static int ixvcom_ioctl(struct tty_struct * tty, unsigned int cmd, unsigned long arg)
#else
static int ixvcom_ioctl(struct tty_struct * tty, struct file * file, unsigned int cmd, unsigned long arg)
#endif
{
	int error;
	struct ixvcom_struct *info = (struct ixvcom_struct *)tty->driver_data;
	struct ixvcom_hwconf *net_dev;
	int port = 0;
        char rsp_buffer[84];
        //char mode_str[16];
        int rsp_length = sizeof(rsp_buffer);
	int retval;
	struct async_icount cprev, cnow;        /* kernel counter temps */
	struct serial_icounter_struct *p_cuser;     /* user space */
	unsigned long templ;
	ixvcom_cmd_t *vcom = (ixvcom_cmd_t *) arg;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (!info)
		return -ENODEV;

        port = info->port;
        net_dev = info->net_dev;

	if ( (cmd != TIOCGSERIAL) && (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT) )
	{
		if ( tty->flags & (1 << TTY_IO_ERROR) )
		{
			return(-EIO);
		}
	}

	switch ( cmd )
	{
		case TCFLSH:
			retval = tty_check_change(tty);
			if (retval)
			{
				return retval;
			}

			switch (arg)
			{
				case TCIFLUSH:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
					if (tty->ldisc->ops->flush_buffer)
						tty->ldisc->ops->flush_buffer(tty);
#else
					if (tty->ldisc.flush_buffer)
						tty->ldisc.flush_buffer(tty);
#endif
					break;

				case TCIOFLUSH:
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
					if (tty->ldisc->ops->flush_buffer)
						tty->ldisc->ops->flush_buffer(tty);
#else
					if (tty->ldisc.flush_buffer)
						tty->ldisc.flush_buffer(tty);
#endif
				case TCOFLUSH:
					ixvcom_flush_buffer(tty);
					break;

				default:
					return -EINVAL;
			}

			return 0;

		case TCSBRK:        /* SVID version: non-zero arg --> no break */
			retval = tty_check_change(tty);

			if ( retval )
				return(retval);

			tty_wait_until_sent(tty, 0);

			if ( !arg )
				ixvcom_send_break(info, HZ/4);	/* 1/4 second */

			return 0;

		case TCSBRKP:       /* support for POSIX tcsendbreak() */
			retval = tty_check_change(tty);

			if ( retval )
				return(retval);

			tty_wait_until_sent(tty, 0);
			ixvcom_send_break(info, arg ? arg*(HZ/10) : HZ/4);
			return 0;

		case TIOCGSOFTCAR:
			error = access_ok(VERIFY_WRITE, (void *)arg, sizeof(long))?0:-EFAULT;

			if ( error )
				return error;

			put_user(C_CLOCAL(tty) ? 1 : 0, (unsigned long *)arg);
			return 0;

		case TIOCSSOFTCAR:
			error = access_ok(VERIFY_READ, (void *)arg, sizeof(long))?0:-EFAULT;
			if ( error )
				return(error);
			
			get_user(templ,(unsigned long *)arg);
			arg = templ;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
                        tty->termios.c_cflag = ((tty->termios.c_cflag & ~CLOCAL) | (arg ? CLOCAL : 0));
#else
                        tty->termios->c_cflag = ((tty->termios->c_cflag & ~CLOCAL) | (arg ? CLOCAL : 0));
#endif
			return 0;

		case TIOCGSERIAL:
			error = access_ok(VERIFY_WRITE, (void *)arg,
			sizeof(struct serial_struct))?0:-EFAULT;

			if ( error )
				return error;

			return(ixvcom_get_serial_info(info, (struct serial_struct *)arg));

		case TIOCSSERIAL:
			error = access_ok(VERIFY_READ, (void *)arg,
			sizeof(struct serial_struct))?0:-EFAULT;

			if ( error )
				return error;

			return(ixvcom_set_serial_info(info, (struct serial_struct *)arg));

		case TIOCSERGETLSR: /* Get line status register */
			error = access_ok(VERIFY_WRITE, (void *)arg,
			sizeof(unsigned int))?0:-EFAULT;

			if ( error )
				return error;
			else
				return (ixvcom_get_lsr_info(info, (unsigned int *)arg));

		case TIOCMIWAIT:
		{
			DECLARE_WAITQUEUE(wait, current);
			int ret;

			cprev = info->icount;   /* note the counters on entry */
			add_wait_queue(&info->delta_msr_wait, &wait);

			while(1)
			{
				/* see if a signal did it */
				cnow = info->icount;	/* atomic copy */
				set_current_state(TASK_INTERRUPTIBLE);

				if ( ((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
				     ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
				     ((arg & TIOCM_CD)   && (cnow.dcd != cprev.dcd)) ||
				     ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts)) )
				{
					ret = 0;
					break;
				}

				if (signal_pending(current))
				{
					ret = -ERESTARTSYS;
					break;
				}

				cprev = cnow;
				schedule();
			}

			set_current_state(TASK_RUNNING);
			remove_wait_queue(&info->delta_msr_wait, &wait);

			return ret;
		}

		case TIOCGICOUNT:
			error = access_ok(VERIFY_WRITE, (void *)arg,
			sizeof(struct serial_icounter_struct))?0:-EFAULT;

			if ( error )
				return(error);

			cnow = info->icount;
			p_cuser = (struct serial_icounter_struct *)arg;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0))
			if (put_user(cnow.frame, &p_cuser->frame))
				return -EFAULT;

			if (put_user(cnow.brk, &p_cuser->brk))
				return -EFAULT;

			if (put_user(cnow.overrun, &p_cuser->overrun))
				return -EFAULT;

			if (put_user(cnow.buf_overrun, &p_cuser->buf_overrun))
				return -EFAULT;

			if (put_user(cnow.parity, &p_cuser->parity))
				return -EFAULT;

			if (put_user(cnow.rx, &p_cuser->rx))
				return -EFAULT;

			if (put_user(cnow.tx, &p_cuser->tx))
				return -EFAULT;
#endif
			put_user(cnow.cts, &p_cuser->cts);
			put_user(cnow.dsr, &p_cuser->dsr);
			put_user(cnow.rng, &p_cuser->rng);
			put_user(cnow.dcd, &p_cuser->dcd);

			return 0;

		case TCXONC:
			retval = tty_check_change(tty);

			if ( retval )
				return retval;

/*
			switch (arg)
			{
				case TCOOFF:
					return ixvcom_setxon_xoff(info, VCOM_CMD_SETXOFF);

				case TCOON:
					return ixvcom_setxon_xoff(info, VCOM_CMD_SETXON);

				default:
					return -EINVAL;
			}
*/
			return 0;

		case IXVCOMCMD:
/*
			if (ixvcom_wait_and_set_command(net_dev, IXVCOM_VCOM_COMMAND_SET, vcom->cmd) < 0)
                        {
                                printk("set VxCOMM.exe command failed\n");
                                return -EIO;
                        }
*/
			//VCOM CMD : [Command][Command Argument]
			sprintf(net_dev->cmd_buffer, "1%s", vcom->cmd);
			net_dev->cmd_ready = 1;

			//printk("XXX %s ... VxCOMM.exe command : %s\n",__FUNCTION__, net_dev->cmd_buffer);

			if (waitqueue_active(&net_dev->select_ex_wait))
				wake_up_interruptible( &net_dev->select_ex_wait );

			if (ixvcom_wait_command_completed(net_dev,
                        	                  IXVCOM_VCOM_COMMAND_SET,
                                	          VCOM_CMD_SET_DATAFORMAT,
                                        	  IXVCOM_CMD_TIMEOUT,
	                                          rsp_buffer,
        	                                  &rsp_length))
			{
				printk("wait VxCOMM.exe command response failed\n");
				return -EIO;
			}

			strncpy(vcom->rsbuf, rsp_buffer, rsp_length);
			//printk("XXX %s ... VxCOMM.exe response : %s\n",__FUNCTION__, rsp_buffer);

		default:
			return -ENOIOCTLCMD;
	}

	return 0;
}

static int ixvcom_open(struct tty_struct * tty, struct file * filp)
{
	struct ixvcom_struct *info;
	int line;
	unsigned long page;
	struct ixvcom_hwconf *net_dev;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	try_module_get(THIS_MODULE);
	line = tty->index;
	//printk("XXX %s ... 1 ... Open Port(%d)  XXX\n",__FUNCTION__, line+1);

	if ( (line < 0) || (line >= VXCOM_MAX_PORTS) )
	{
		printk("line %d : over range\n", line);
		module_put(THIS_MODULE);
		return -ENODEV;
	}

	info = ixvcom_table + line;

	net_dev = info->net_dev;

	if ( !net_dev  || !(net_dev->flag & IXVCOM_NET_DEV_OPENED))
	{
		printk("Net device isn't opened\n");
		module_put(THIS_MODULE);
		return -ENODEV;
	}

	if ( !ixvcom_tmp_buf )
	{
		page = __get_free_page(GFP_KERNEL);

		if ( !page )
		{
			printk("ixvcom_tmp_buf get free page failed\n");
			module_put(THIS_MODULE);
			return -ENOMEM;
		}

		if ( ixvcom_tmp_buf )
			free_page(page);
		else
			ixvcom_tmp_buf = (unsigned char *)page;
	}

	/*Start up serial port*/
	tty->driver_data = info;
	info->count++;

	if (ixvcom_startup(info, filp, tty))
	{
		printk("ixvcom startup serial port failed\n");
		return -EIO;
	}

	if (ixvcom_block_til_ready(tty, filp, info))
	{
		printk("ixvcom_block_til_ready failed\n");
		info->count--;
		module_put(THIS_MODULE);
		return -EIO;
	}

	if ( (info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS) )
	{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
                if ( tty->driver->subtype == SERIAL_TYPE_NORMAL )
                        tty->termios = info->normal_termios;
                else
                        tty->termios = info->callout_termios;
#else
                if ( tty->driver->subtype == SERIAL_TYPE_NORMAL )
                        *tty->termios = info->normal_termios;
                else
                        *tty->termios = info->callout_termios;
#endif

/*
		if (ixvcom_port_init(info, 0))
		{
			printk("ixvcom port init failed\n");
			return -EIO;
		}
*/
	}

	info->session = MX_SESSION();
	info->pgrp = MX_CGRP();
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,9,0))
        tty->port->low_latency = 1;
#else
        tty->low_latency = 1;
#endif
	//printk("SSS %s ... END SSS\n",__FUNCTION__);

	return 0;
}

static void ixvcom_close(struct tty_struct * tty, struct file * filp)
{
	struct ixvcom_struct * info = (struct ixvcom_struct *)tty->driver_data;
	long timeout, et, ret;
	int cnt;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if ( !info )
		return;

	if ( tty_hung_up_p(filp) )
	{
		info->count--;
		module_put(THIS_MODULE);
		return;
	}

#ifndef SP1
	if ( (tty->count == 1) && (info->count != 1) )
	{
#else
        if ( (atomic_read(&tty->count) == 1) && (info->count != 1) )
        {
#endif
		printk("[%d] ixvcom_close: bad serial port count; tty->count is 1, info->count is %d\n", current->pid, info->count);
		info->count = 1;
	}

	if ( --info->count < 0 )
	{
		printk("ixvcom_close: bad serial port count for port %d: %d\n", info->port, info->count);
		info->count = 0;
	}

	if ( info->count )
	{
		module_put(THIS_MODULE);
		return;
	}

	 if (info->flags & ASYNC_CLOSING)
		return;

	info->flags |= ASYNC_CLOSING;
	tty->closing = 1;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
        if ( info->flags & ASYNC_NORMAL_ACTIVE )
                info->normal_termios = tty->termios;

        if ( info->flags & ASYNC_CALLOUT_ACTIVE )
                info->callout_termios = tty->termios;
#else
        if ( info->flags & ASYNC_NORMAL_ACTIVE )
                info->normal_termios = *tty->termios;

        if ( info->flags & ASYNC_CALLOUT_ACTIVE )
                info->callout_termios = *tty->termios;
#endif

	if ( !(filp->f_flags & O_NONBLOCK) && info->closing_wait != ASYNC_CLOSING_WAIT_NONE )
	{
		et = jiffies + info->closing_wait;
		tty_wait_until_sent(tty, info->closing_wait);
		cnt = 0;

		while ((timeout = et - jiffies) > 0)
		{
			if ((ret = ixvcom_wait_oqueue(info, timeout)) == 0)
			{
				if (++cnt >= 3)
				{
					break;
				}
			}
			else if (ret < 0)
			{
				break;
			}
			else
			{
				cnt = 0;
			}

			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout(HZ/100);
		}
	}

	ixvcom_flush_buffer(tty);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
	if ( tty->ldisc->ops->flush_buffer )
		tty->ldisc->ops->flush_buffer(tty);
#else
	if ( tty->ldisc.flush_buffer )
		tty->ldisc.flush_buffer(tty);
#endif

	ixvcom_shutdown(info);
	tty->closing = 0;
	module_put(THIS_MODULE);

	//printk("SSS %s ... END SSS\n",__FUNCTION__);
}


#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
static void ixvcom_flush_to_ldisc(void *private_)
{
	struct ixvcom_struct *info = (struct ixvcom_struct *)private_;
#else
static void ixvcom_flush_to_ldisc(struct work_struct *work)
{
	struct ixvcom_struct *info = container_of(work, struct ixvcom_struct, process_flip_tqueue);
#endif
	struct tty_struct *tty;
	int count;

#if ((LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) || ((LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15)) && defined(FEDORA)))
	struct tty_ldisc *disc;
	struct tty_buffer *tbuf, *head;
	unsigned long       flags;
	unsigned char       *fp;
	char                *cp;
#else
	unsigned char       *cp;
	char                *fp;
#endif

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	if (info)
	{
#if ((LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,15)) && !defined(FEDORA))
		down(&info->rx_semaphore);
#endif
		tty = info->tty;
//
#if ((LINUX_VERSION_CODE > KERNEL_VERSION(2,6,15)) || ((LINUX_VERSION_CODE == KERNEL_VERSION(2,6,15)) && defined(FEDORA)))
		disc = tty_ldisc_ref(tty);

		if (disc == NULL)   
			return;
#endif

		if ( tty && (info->flags & ASYNC_INITIALIZED))
		{
#if ((LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,15)) && !defined(FEDORA))
			if (tty->flip.buf_num)
			{
				cp = tty->flip.char_buf + TTY_FLIPBUF_SIZE;
				fp = tty->flip.flag_buf + TTY_FLIPBUF_SIZE;
				tty->flip.buf_num = 0;
				tty->flip.char_buf_ptr = tty->flip.char_buf;
				tty->flip.flag_buf_ptr = tty->flip.flag_buf;
			}
			else
			{
				cp = tty->flip.char_buf;
				fp = tty->flip.flag_buf;
				tty->flip.buf_num = 1;

				tty->flip.char_buf_ptr = tty->flip.char_buf + TTY_FLIPBUF_SIZE;
				tty->flip.flag_buf_ptr = tty->flip.flag_buf + TTY_FLIPBUF_SIZE;
			}

			count = tty->flip.count;
			tty->flip.count = 0;
			tty->ldisc.receive_buf(tty, cp, fp, count);
#else

			#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
                                spin_lock_irqsave(&tty->port->buf.lock, flags);
                                head = tty->port->buf.head;
                        #else
                                spin_lock_irqsave(&tty->buf.lock, flags);
                                head = tty->buf.head;
                        #endif

			if (head != NULL)
			{
				#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
                                        tty->port->buf.head = NULL;
                                #else
                                        tty->buf.head = NULL;
                                #endif

				for (;;)
				{
					count = head->commit - head->read;

					if (!count)
					{
						if (head->next == NULL)
							break;

						tbuf = head;
						head = head->next;
						//tty_buffer_free(tty);
						tty_buffer_free(tty, tbuf);
						continue;
					}

					if (!tty->receive_room)
					{
#if ((LINUX_VERSION_CODE < KERNEL_VERSION(2,6,39)) && !defined(FEDORA))
						schedule_delayed_work(&tty->buf.work, 1);
#else
						schedule_work(&tty->SAK_work);
#endif

/*
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
						schedule_delayed_work(&(tty->port->buf.work), 1);
#else
						schedule_delayed_work(&tty->buf.work, 1);
#endif
*/	
						break;
					}

					if (count > tty->receive_room)
					{
						count = tty->receive_room;
					}

					cp = head->char_buf_ptr + head->read;
					fp = head->flag_buf_ptr + head->read;
					head->read += count;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
                                        spin_unlock_irqrestore(&tty->port->buf.lock, flags);
#else
                                        spin_unlock_irqrestore(&tty->buf.lock, flags);
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,27))
					disc->ops->receive_buf(tty, cp, fp, count);
#else
					disc->receive_buf(tty, cp, fp, count);
#endif

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
                                        spin_lock_irqsave(&tty->port->buf.lock, flags);
#else
                                        spin_lock_irqsave(&tty->buf.lock, flags);
#endif
				}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
                                tty->port->buf.head = head;
#else
                                tty->buf.head = head;
#endif
			}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
                        spin_unlock_irqrestore(&tty->port->buf.lock, flags);
#else
                        spin_unlock_irqrestore(&tty->buf.lock, flags);
#endif
			tty_ldisc_deref(disc);
#endif
		}

#if ((LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,15)) && !defined(FEDORA))
		up(&info->rx_semaphore);
#endif
	}
}


static int __init ixvcom_init_tty(void)
{
	struct ixvcom_struct *tty_node;
	int i;
	struct proc_dir_entry *pdir;
	struct ixvcom_hwconf *net_dev;
	char buf[4];

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
	sema_init(&ixvcom_tmp_buf_sem, 1);
#else
	init_MUTEX(&ixvcom_tmp_buf_sem);
#endif


	//create ixvcom proc directory
	ixvcom_proc_dir = proc_mkdir(FAMILY, 0);

	if ( !ixvcom_proc_dir  )
		return -ENOMEM;

	tty_node = &ixvcom_table[0];
	net_dev = &ixvcom_net_dev[0];

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
        if(!proc_create(DEVICE_NAME, 0, ixvcom_proc_dir, &ixvcom_proc_fops))
        {
                printk("IxVCOM : unable to register /proc/ixvcom/ixvcom\n");
                return -ENOMEM;
        }
#else
        create_proc_read_entry(DEVICE_NAME, 0, ixvcom_proc_dir, ixvcom_read_proc, NULL);
#endif

	for ( i = 0; i < VXCOM_MAX_PORTS; i++, tty_node++, net_dev++ )
	{
		sprintf(buf, "%d", i);

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,8,0))
                pdir = proc_create_data(buf, S_IRUGO | S_IWUGO | S_IFREG, ixvcom_proc_dir, &ixvcom_net_fops, (void *) net_dev);
                if(!pdir)
                {
                        printk("IxVCOM : unable to register /proc/ixvcom/%d\n", i);
                        return -ENOMEM;
                }
#else
                pdir = create_proc_entry( buf, S_IRUGO | S_IWUGO | S_IFREG, ixvcom_proc_dir);
#endif

		if ( !pdir )
			return -ENOMEM;


#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0))
		pdir->data = (void *) net_dev;
		pdir->proc_fops = &ixvcom_net_fops;
#endif

		net_dev->tty_node = tty_node;
		net_dev->node_entry = pdir;
		net_dev->flag = 0;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
		sema_init(&net_dev->semaphore, 1);
		sema_init(&net_dev->cmd_semaphore, 1);
#else
		init_MUTEX(&net_dev->semaphore);
		init_MUTEX(&net_dev->cmd_semaphore);
#endif
		init_waitqueue_head(&net_dev->initialize_wait);
		init_waitqueue_head(&net_dev->select_in_wait);
		init_waitqueue_head(&net_dev->select_out_wait);
		init_waitqueue_head(&net_dev->select_ex_wait);
		init_waitqueue_head(&net_dev->cmd_rsp_wait);

		tty_node->net_dev = net_dev;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0))
		tty_node->port = 0;
#else
		tty_node->port = i;
#endif
		tty_node->uart_type = PORT_16550A;
		tty_node->flags = 0;
		tty_node->xmit_fifo_size = 16;
		tty_node->baud_base = 921600L;
		tty_node->close_delay = 5*HZ/10;
		tty_node->closing_wait = 30*HZ;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20))
		INIT_WORK(&tty_node->tqueue, ixvcom_do_softint, tty_node);
		INIT_WORK(&tty_node->process_flip_tqueue, ixvcom_flush_to_ldisc, tty_node);
#else
		INIT_WORK(&tty_node->tqueue, ixvcom_do_softint);
		INIT_WORK(&tty_node->process_flip_tqueue, ixvcom_flush_to_ldisc);
#endif

		//tty_node->normal_termios = DRV_VAR_P(init_termios);
		tty_node->normal_termios = ixvcom_driver->init_termios;

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36))
		tty_node->tx_lock = __SPIN_LOCK_UNLOCKED(tty_node->tx_lock);
		//__SPIN_LOCK_UNLOCKED(tty_node->tx_lock);
		init_MUTEX(&tty_node->rx_semaphore);
		//sema_init(&tty_node->rx_semaphore, 1);
#else
		tty_node->tx_lock = SPIN_LOCK_UNLOCKED;
		init_MUTEX(&tty_node->rx_semaphore);
#endif
		init_waitqueue_head(&tty_node->open_wait);
		init_waitqueue_head(&tty_node->close_wait);
		init_waitqueue_head(&tty_node->delta_msr_wait);

		tty_node->icount.rx = tty_node->icount.tx = 0;
		tty_node->icount.cts = tty_node->icount.dsr = tty_node->icount.dsr =  tty_node->icount.dcd = 0;

		tty_node->icount.frame = tty_node->icount.overrun = tty_node->icount.brk = tty_node->icount.parity = 0;
	}

	return 0;
}

static int __init ixvcom_init(void)
{
	int retval = 0;
	int i;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	ixvcom_driver = alloc_tty_driver(VXCOM_MAX_PORTS + 1);

	if (!ixvcom_driver)
		return -ENOMEM;
#endif

	printk("ICP DAS Virtual Driver Version %s\n", IXVCOM_DRIVER_VERSION);

	/* tty_driver structure init*/
	ixvcom_driver->magic = TTY_DRIVER_MAGIC;
        ixvcom_driver->owner = THIS_MODULE;
        ixvcom_driver->driver_name = "ixvcom";
        ixvcom_driver->name = "ttyVCOM";
        ixvcom_driver->major = 0;
        ixvcom_driver->minor_start = 0;
	ixvcom_driver->num = VXCOM_MAX_PORTS;
	//ixvcom_driver->name_base = 0;
        ixvcom_driver->type = TTY_DRIVER_TYPE_SERIAL;
        ixvcom_driver->subtype = SERIAL_TYPE_NORMAL;
        ixvcom_driver->init_termios = tty_std_termios;
        ixvcom_driver->init_termios.c_cflag = B9600 | CS8 | CREAD | HUPCL | CLOCAL;

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,18))
	ixvcom_driver->flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS;
#else
	ixvcom_driver->flags = TTY_DRIVER_REAL_RAW |  TTY_DRIVER_DYNAMIC_DEV;
	//ixvcom_driver->flags = TTY_DRIVER_REAL_RAW | TTY_DRIVER_DYNAMIC_DEV;
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0))
        ixvcom_driver->ttys = ixvcom_tty;
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,7,0))
        ixvcom_driver->termios = ixvcom_termios;
#endif


#if (LINUX_VERSION_CODE <= KERNEL_VERSION(3,2,20))
        ixvcom_driver->termios_locked = ixvcom_termios_locked;
#endif
	tty_set_operations(ixvcom_driver, &ixvcom_ops);

	memset(ixvcom_table, 0, VXCOM_MAX_PORTS * sizeof(struct ixvcom_struct));
	memset(&ixvcom_net_dev, 0, sizeof(struct ixvcom_hwconf));

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
	for(i = 0; i < VXCOM_MAX_PORTS;i++)
	{
		tty_port_init(&ixvcom_table[i].ttyPort);
		tty_port_link_device(&ixvcom_table[i].ttyPort, ixvcom_driver, i);
	}
#endif
	/*register driver*/

	retval = tty_register_driver(ixvcom_driver);

	if (!retval)
        {
		/* get major number */
		//printk("ixvcom_major : %d\n", ixvcom_driver->major);
		ixvcom_major = ixvcom_driver->major;
        }
        else
        {
		printk(KERN_ERR "Couldn't install ICP DAS Virtual COM Driver !\n");
		put_tty_driver(ixvcom_driver);

		return retval;
        }

	/* init ixvcom_net_fops structure init*/

	memset(&ixvcom_net_fops, 0, sizeof(struct file_operations));
	ixvcom_net_fops.read = ixvcom_net_read;
	ixvcom_net_fops.write = ixvcom_net_write;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,36)
	ixvcom_net_fops.unlocked_ioctl = ixvcom_net_ioctl;
#else
	ixvcom_net_fops.ioctl = ixvcom_net_ioctl;
#endif
	ixvcom_net_fops.open = ixvcom_net_open;
	ixvcom_net_fops.release = ixvcom_net_close;
	ixvcom_net_fops.poll = ixvcom_net_select;

	if (ixvcom_init_tty() != 0)
	{
		tty_unregister_driver(ixvcom_driver);
		printk(KERN_ERR "Couldn't install ICP DAS Virtual COM Driver !\n");
	}
	
	printk("%s : Install ICP DAS Virtual COM Driver OK!\n", __FUNCTION__);

	return 0;
}

static void __exit ixvcom_exit(void)
{
	int i;
	int retval = 0;
	struct ixvcom_struct *info;
	struct proc_dir_entry *pdir;

	//printk("SSS %s ... START SSS\n",__FUNCTION__);

	info = &ixvcom_table[0];

	for (i = 0; i < VXCOM_MAX_PORTS; i++, info++)
	{
		if (info->net_dev)
		{
			if ((pdir = ((struct ixvcom_hwconf *)(info->net_dev))->node_entry))
			{
				if(pdir)
				{
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0))
                                        proc_remove(pdir);
#else
                                        remove_proc_entry(pdir->name, pdir->parent);
#endif
				}
			}

			#if (LINUX_VERSION_CODE >= KERNEL_VERSION(3,7,0))
			/* Remove device node if it is opened*/
			if(info->net_dev->flag & IXVCOM_NET_DEV_OPENED)
				tty_unregister_device(ixvcom_driver, i);
			#endif


			((struct ixvcom_hwconf *)(info->net_dev))->node_entry = NULL;
		}
	}

	if (ixvcom_proc_dir)
	{
		remove_proc_entry(DEVICE_NAME, ixvcom_proc_dir);
		remove_proc_entry(FAMILY, NULL);
		ixvcom_proc_dir = NULL;
	}

	retval = tty_unregister_driver(ixvcom_driver);

	if (retval)
	{
		printk(KERN_ERR "Couldn't remove ICP DAS Virtual COM Driver !\n");
	}

	put_tty_driver(ixvcom_driver);

	printk("%s : Remove ICP DAS Virtual COM Driver OK!\n", __FUNCTION__);
}

module_init(ixvcom_init);
module_exit(ixvcom_exit);

